# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import asyncio
import traceback
from typing import TYPE_CHECKING, Awaitable, Callable, List, Optional

import aiohttp
from natsort import natsorted

from johann import api_conductor as api
from johann.shared.config import JohannConfig
from johann.shared.logger import JohannLogger
from johann.util import johann_response

if TYPE_CHECKING:
    from aiohttp.web import Request, Response

config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


@aiohttp.web.middleware
async def error_middleware(
    request: "Request", handler: Callable[["Request"], Awaitable]
) -> "Response":
    try:
        response = await handler(request)
        return response
    except aiohttp.web.HTTPError as e:
        return johann_response(False, str(e), e.status_code)
    except Exception:
        msg = (
            "Error handling API request"
            f" ({handler.__name__}):\n{traceback.format_exc()}"
        )
        logger.error(msg)
        if config.DEBUG:
            return aiohttp.web.Response(body=msg, status=500)
        else:
            return johann_response(False, "internal error", 500)


def get_routes(path_substr: Optional[str] = None) -> List:
    ret = []

    for r in app.router.routes():
        ri = r.get_info()
        rip = None
        if "path" in list(ri.keys()):
            rip = ri["path"]
        elif "formatter" in list(ri.keys()):
            rip = ri["formatter"]

        if rip and (not path_substr or path_substr in rip) and rip not in ret:
            ret.append(rip)

    return natsorted(ret)


async def api_get_routes(request: "Request") -> "Response":
    if config.DEBUG:
        logger.debug(f"{request.url}")
    return johann_response(True, [], data=get_routes())


def init_conductor() -> None:
    # routes
    app.router.add_get("/", api_get_routes)
    app.router.add_get("/routes", api_get_routes)
    app.router.add_get("/codehash", api.api_get_codehash)
    app.router.add_routes(
        [
            aiohttp.web.get("/affrettando/{score_name}", api.affrettando),
            aiohttp.web.post("/affrettando/{score_name}", api.affrettando),
        ]
    )
    app.router.add_get("/read_score/{score_name}", api.api_read_score)
    app.router.add_get("/scores", api.get_scores)
    app.router.add_get("/scores/{score_name}", api.get_score)
    app.router.add_get("/scores/{score_name}/get_raw", api.get_score_raw)
    app.router.add_get("/scores/{score_name}/status", api.get_score_status)
    app.router.add_get("/scores/{score_name}/status_short", api.get_score_status_short)
    app.router.add_get("/scores/{score_name}/status_alt", api.get_score_status_alt)
    app.router.add_get("/scores/{score_name}/measures", api.get_score_measures)
    app.router.add_get("/scores/{score_name}/measures/{measure_name}", api.get_measure)
    app.router.add_get(
        "/scores/{score_name}/measures/{measure_name}/status", api.get_measure_status
    )
    app.router.add_get(
        "/scores/{score_name}/measures/{measure_name}/play", api.manually_play_measure
    )
    app.router.add_get("/scores/{score_name}/stored_data", api.retrieve_stored_data_all)
    app.router.add_get(
        "/scores/{score_name}/stored_data/{key}", api.retrieve_stored_data_1
    )
    app.router.add_get(
        "/scores/{score_name}/stored_data/{key}/{subkey}", api.retrieve_stored_data_2
    )
    app.router.add_get(
        "/scores/{score_name}/stored_data/{key}/{subkey}/{subsubkey}",
        api.retrieve_stored_data_3,
    )
    app.router.add_get(
        "/scores/{score_name}/stored_data/{key}/{subkey}/{subsubkey}/{subsubsubkey}",
        api.retrieve_stored_data_4,
    )
    app.router.add_routes(
        [
            aiohttp.web.get("/scores/{score_name}/roll_call", api.roll_call),
            aiohttp.web.post("/scores/{score_name}/roll_call", api.roll_call),
        ]
    )
    app.router.add_get("/scores/{score_name}/cue_the_music", api.cue_the_music)
    app.router.add_get("/hosts", api.get_hosts)
    app.router.add_get("/hosts/{host_name}", api.get_host)
    app.router.add_post("/add_hosts", api.add_hosts)
    # app.router.add_get('/create_player/{score_name}/{player_name}', api.api_create_player)

    logger.info("********** Reading Score Files **********")
    api.read_scores()

    logger.info("********** Adding Initial Hosts **********")
    api._update_hosts(config.INITIAL_HOSTS, allow_invalid=True)

    if config.HOSTS_FILE:
        logger.info("********** Reading Hosts File **********")
        api.read_hosts_file(config.HOSTS_FILE)

    logger.info("********** Starting Web Server **********")

    runner = aiohttp.web.AppRunner(app)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(runner.setup())
    site = aiohttp.web.TCPSite(
        runner, str(config.JOHANN_HOST.ip), config.CONDUCTOR_PORT
    )
    loop.run_until_complete(site.start())


app = aiohttp.web.Application(middlewares=[error_middleware])



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.

"""Johann, lightweight and flexible scenario orchestration"""

__version__ = "0.2.1-alpha"



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
from typing import TYPE_CHECKING, List, Optional, Tuple, Type

from johann.docker_host_control import DockerHostControl
from johann.shared.config import JohannConfig
from johann.shared.logger import JohannLogger
from johann.util import get_attr

if TYPE_CHECKING:
    from johann.host_control import HostControl


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


def get_host_control_class(
    control_method: str,
) -> Tuple[Optional[Type["HostControl"]], Optional[str]]:
    if control_method.upper() == "DOCKER":
        return DockerHostControl, None
    elif control_method not in config.HOST_CONTROL_CLASS_NAMES:
        msg = f"unrecognized control method '{control_method}'"
        return None, msg

    host_control_class, msg = get_attr(config.HOST_CONTROL_CLASS_NAMES[control_method])
    if not host_control_class:
        msg = f"failed to find class for control method '{control_method}'"
        return None, msg

    return host_control_class, None


def get_host_names(
    control_methods: Optional[List[str]] = None,
) -> List[str]:
    if not control_methods:
        control_methods = config.HOST_CONTROL_CLASS_NAMES.keys()
    ret = []
    for cm in control_methods:
        cm_class, msg = get_host_control_class(cm)
        if not cm_class:
            logger.error(msg)
            continue
        try:
            cm_hosts = cm_class.get_host_names()
        except AttributeError as e:
            logger.error(f"{cm_class}: {str(e)}")
        else:
            ret += cm_hosts
            logger.debug(f"Found {len(cm_hosts)} hosts via {cm}")

    return ret



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import asyncio
import copy
import pprint
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple
from uuid import uuid4

from marshmallow import Schema
from marshmallow import ValidationError as MarshmallowValidationError
from marshmallow import fields, post_load

from johann.docker_host_control import DockerHostControl
from johann.host import HostSchema
from johann.host_control_util import get_host_control_class, get_host_names
from johann.measure import MeasureSchema
from johann.player import PlayerSchema
from johann.shared.config import JohannConfig, hosts, scores
from johann.shared.enums import TaskState
from johann.shared.fields import NameField, StateField
from johann.shared.logger import JohannLogger
from johann.util import (
    create_johann_tarball,
    gudexc,
    gudlog,
    task_state_priority,
    transform_arg,
    transform_args,
    wrap_future,
)

if TYPE_CHECKING:
    from johann.host import Host
    from johann.measure import Measure
    from johann.player import Player


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


conductor_local_measure_dicts = [
    {
        "name": "tune_orchestra",
        "players": [config.CONDUCTOR_ALLHOSTS_PLAYER_NAME],
        "task": "johann.conductor_tasks.tune_host",
        "start_delay": 0,
        # 'depends_on': ['create_hosts']
    },
    # {
    # 'name': 'create_hosts',
    # 'players': [config.CONDUCTOR_ALLHOSTS_PLAYER_NAME],
    # 'task': 'johann.conductor_tasks.create_hosts',
    # 'start_delay': 10
    # },
]


class ScoreSchema(Schema):
    class Meta:
        ordered = True

    name = NameField(required=True)
    version = fields.Str(missing="1.0")
    category = fields.Str(missing="none")
    description = fields.Str(missing="")
    players = fields.Dict(
        keys=fields.Str(), values=fields.Nested(PlayerSchema), required=True
    )
    measures = fields.List(fields.Nested(MeasureSchema), required=True)
    create_hosts = fields.Boolean(missing=False)
    discard_hosts = fields.Boolean(missing=False)

    # do not remove any of these three without careful consideration
    state = StateField(dump_only=True)
    status = fields.Dict(
        keys=fields.Str(),  # measure
        values=fields.Dict(
            keys=fields.Str(),  # player
            values=fields.Dict(
                keys=fields.Str(), values=fields.Str()  # host, host status
            ),
        ),
        dump_only=True,
    )
    finished = fields.Boolean(dump_only=True)

    package = fields.Str(dump_only=True)
    started_at = fields.DateTime(dump_only=True)
    finished_at = fields.DateTime(dump_only=True)
    stored_data = fields.Dict(keys=fields.Str(), values=fields.Raw(), dump_only=True)
    task_map = fields.Dict(
        keys=fields.Str(),
        values=fields.Dict(keys=fields.Str(), values=fields.Str()),
        dump_only=True,
    )
    last_successful_roll_call = fields.DateTime(dump_only=True)

    # depends on post_load, so we do not use @validates_schema here
    def validate_score(self, score: "Score", **kwargs) -> None:
        self.validate_unique(score)
        self.validate_measure_player_names(score)
        self.validate_measure_depends_on(score)

    def validate_unique(self, score: "Score", **kwargs) -> None:
        if score.name in scores:
            raise MarshmallowValidationError(
                f"There is already a score with name '{score.name}'"
            )

    def validate_measure_player_names(self, score: "Score", **kwargs) -> None:
        # no duplicate measure names
        measure_names = [x.name for x in score.measures]
        if len(set(measure_names)) != len(measure_names):
            raise MarshmallowValidationError("Duplicate measure name(s)")

        for measure in score.measures:
            # all players specified in measure also exist in score's list of players
            for player_name in measure.player_names:
                if player_name not in score.players:
                    raise MarshmallowValidationError(
                        f"Unrecognized player '{player_name}' in measure"
                        f" '{measure.name}'"
                    )

    def validate_measure_depends_on(self, score: "Score", **kwargs) -> None:
        measure_names = [x.name for x in score.measures]
        bad_deps = []
        for measure in score.measures:
            for dep in measure.depends_on:
                if dep not in measure_names:
                    bad_deps.append(dep)

        if len(bad_deps) > 0:
            raise MarshmallowValidationError(
                f"Invalid values for depends_on:\n{bad_deps}"
            )

    @post_load(pass_original=True)
    def make_score(self, data: Dict, original_data: Dict, **kwargs) -> "Score":
        score = Score(**data, original_data=original_data)
        self.validate_score(score)
        return score


class Score(object):
    def __init__(
        self,
        name,
        version,
        category,
        description,
        players,
        measures,
        create_hosts,
        discard_hosts,
        original_data,
    ) -> None:
        self.name: str = name
        self.version: str = version
        self.category: str = category
        self.description: str = description
        self.players: Dict[str, "Player"] = players
        self.measures: List["Measure"] = measures
        self.create_hosts: bool = create_hosts
        self.discard_hosts: bool = discard_hosts
        self.original_data: Dict[str, Any] = original_data

        self.state: TaskState = TaskState.PENDING
        self.status: Dict[str, Dict[str, Dict[str, str]]] = {}
        self.package: Optional[str] = None
        self.finished: bool = False
        self.started_at: Optional[datetime] = None
        self.finished_at: Optional[datetime] = None
        self.stored_data: Dict[str, Any] = {}
        self.task_map: Dict[
            str, Dict[str, str]
        ] = {}  # {<TASK_ID>: {"measure_name":<>, "player_name":<>, "host_name":<>}}
        # this is used to map tasks within a celery group back to the
        # player/host that they 'belong' to
        self.last_successful_roll_call: Optional[datetime] = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self.name},version={self.version},"
            f"category={self.category},description={self.description},"
            f"players={'|'.join(list(self.players.keys()))},"
            f"measures={'|'.join([m.name for m in self.measures])})"
        )

    def dump(
        self, exclude_local: bool = True, yaml_fields_only: bool = True
    ) -> Dict[str, Any]:
        data = ScoreSchema().dump(self)

        if exclude_local:
            # remove local player
            try:
                del data["players"][config.CONDUCTOR_ALLHOSTS_PLAYER_NAME]
            except KeyError:
                pass

            local_measure_names = [x["name"] for x in conductor_local_measure_dicts]
            to_remove = []
            for m in data["measures"]:
                # remove local measure dependencies
                for local_name in local_measure_names:
                    if local_name in m["depends_on"]:
                        m["depends_on"].remove(local_name)
                # find local measures to remove
                if m["name"] in local_measure_names:
                    to_remove.append(m)
            # actually remove the measures (no modify while iterating)
            for m in to_remove:
                data["measures"].remove(m)

        if yaml_fields_only:
            # remove score-level fields
            dump_only_fields = []
            for k, v in ScoreSchema().fields.items():
                if "dump_only=True" in str(v):
                    dump_only_fields.append(k)
            for f in dump_only_fields:
                del data[f]

            # remove measure-level fields
            dump_only_fields = []
            for k, v in MeasureSchema().fields.items():
                if "dump_only=True" in str(v):
                    dump_only_fields.append(k)
            for m in data["measures"]:
                for f in dump_only_fields:
                    del m[f]

        return data

    def get_status(self, short: bool = False) -> Dict[str, Any]:
        status = {"state": self.state, "finished": self.finished}
        if self.status or not short:
            status["status"] = self.status

        if not short:
            status["started_at"] = self.started_at
            status["finished_at"] = self.finished_at
            if self.started_at is not None:
                status["started_at"] = self.started_at.isoformat()
            if self.finished_at is not None:
                status["finished_at"] = self.finished_at.isoformat()
            if self.last_successful_roll_call is not None:
                status[
                    "last_successful_roll_call"
                ] = self.last_successful_roll_call.isoformat()

        status["measures"] = {x.name: x.get_status(short=short) for x in self.measures}

        if not short:
            status["stored_data"] = self.stored_data

        return status

    def get_host_totals(self) -> Dict[str, Any]:
        ret = {"total": 0}
        for m in self.measures:
            ret[m.name] = {}
            ret[m.name]["total"] = 0
            for p_name in m.player_names:
                p = self.players[p_name]
                subtotal = p.scale
                ret[m.name][p.name] = subtotal
                ret[m.name]["total"] += subtotal
                ret["total"] += subtotal
        return ret

    def get_measure(self, name: str) -> Optional["Measure"]:
        for m in self.measures:
            if m.name == name:
                return m
        return None

    # param taken_hostnames should be supplied by host_control.get_host_names()
    def map_missing_hosts(self, taken_hostnames: List[str]) -> None:
        for player in self.players.values():
            count = player.scale - len(player.hostnames)
            for _ in range(count):
                hostname = f"{player.name}_{str(uuid4())[:6]}"
                while hostname in taken_hostnames:  # no duplicates
                    hostname = f"{player.name}_{str(uuid4())[:6]}"
                player.hostnames.append(hostname)
                taken_hostnames.append(hostname)

        return

    def queue_measure(self, measure: "Measure") -> None:
        measure.state = TaskState.QUEUED
        logger.debug(f"Queueing measure {measure.name}")
        for player_name in measure.player_names:
            task = asyncio.ensure_future(
                self.play_the_player(measure, self.players[player_name])
            )
            task = wrap_future(
                task,
                f"{self.name}.{player_name}.{measure.name}",
                None,
                score_or_measure=measure,
            )
            asyncio.ensure_future(task)

    def get_all_hosts(self) -> Tuple[Set["Host"], List[str]]:
        all_hosts = set()
        error_msgs = []
        for p in self.players.values():
            if p.name == config.CONDUCTOR_ALLHOSTS_PLAYER_NAME:
                continue
            for hostname in p.hostnames:
                if hostname not in hosts:
                    error_msgs.append(f"host '{hostname}' not found in orchestra")
                    continue
                else:
                    all_hosts.add(hosts[hostname])

        return all_hosts, error_msgs

    def evaluate_state(self) -> None:
        # make sure we are current on measure state
        for m in self.measures:
            if not m.finished:
                m.evaluate_state(self)

        not_success = 0
        unfinished = 0
        for m in self.measures:
            if not m.finished:
                unfinished += 1
            if m.state is not TaskState.SUCCESS:
                not_success += 1

            if task_state_priority(m.state) > task_state_priority(self.state):
                self.state = m.state

            if m.status:

                # replace task id with host name if known
                status_copy = copy.copy(m.status)
                for player_name, player_status in m.status.items():
                    new_player_status = {}
                    for task_id, status in player_status.items():
                        if task_id in self.task_map:
                            tmi = self.task_map[task_id]
                            new_player_status[tmi["host_name"]] = status
                        else:
                            new_player_status[task_id] = status
                    status_copy[player_name] = new_player_status

                self.status[m.name] = status_copy

        if not_success == 0:
            self.state = TaskState.SUCCESS

        if unfinished == 0:
            self.finished = True
            self.finished_at = datetime.utcnow()
            msg = f"finished with state {self.state}"
            logger.info(gudlog(msg, self))

    def smart_queue(self, unqueued_measures: List["Measure"]) -> Optional["Measure"]:
        to_queue = (
            None  # queue one at a time to ensure we catch any dependency failures
        )
        dependency_failed = []
        for m in unqueued_measures:
            # no dependencies
            if not m.depends_on:
                to_queue = m
                break
            else:
                # make sure dependencies finished before queueing
                # also if any dependencies failed, dependent measures fail too
                ready_to_queue = True  # until/unless False below

                for dep_name in m.depends_on:
                    depm = self.get_measure(dep_name)
                    if not depm.finished:
                        ready_to_queue = False
                        m.state = TaskState.DEFERRED

                    if depm.state == TaskState.FAILURE and not m.dependency_proof:
                        ready_to_queue = False
                        m.state = TaskState.FAILURE
                        m.finished = True
                        m.status["all"]["all"] = f"dependency failed ({dep_name})"
                        dependency_failed.append(m)

                if ready_to_queue:
                    to_queue = m
                    break

        if to_queue is not None:
            self.queue_measure(to_queue)
            if to_queue in unqueued_measures:
                unqueued_measures.remove(to_queue)

        for to_remove in dependency_failed:
            msg = (
                f"measure '{to_remove.name}' failed because one or more of its"
                " dependencies failed"
            )
            logger.warning(gudlog(msg, self))
            if to_remove in unqueued_measures:
                unqueued_measures.remove(to_remove)

        return to_queue

    async def play(self) -> None:
        # add local measures
        for md in conductor_local_measure_dicts:
            if md["name"] not in [
                x.name for x in self.measures
            ]:  # make sure it's not already there
                local_measure: "Measure" = MeasureSchema().load(md)
                local_measure.local_measure = True
                self.measures.insert(0, local_measure)

        # add local measure dependencies
        local_measure_names = [x["name"] for x in conductor_local_measure_dicts]
        for m in self.measures:
            if m.name in local_measure_names:
                continue

            if "tune_orchestra" not in m.depends_on and m.name != "tune_orchestra":
                msg = "adding 'tune_orchestra' dependency"
                logger.debug(gudlog(msg, self, None, m.name))
                m.depends_on.insert(0, "tune_orchestra")

        # add conductor allhosts player
        if config.CONDUCTOR_ALLHOSTS_PLAYER_NAME not in self.players.keys():
            local_player: "Player" = PlayerSchema().load(
                {"name": config.CONDUCTOR_ALLHOSTS_PLAYER_NAME}
            )
            self.players[config.CONDUCTOR_ALLHOSTS_PLAYER_NAME] = local_player

        # conductor player will operate over all hosts, including those not created yet
        all_hosts, err_msgs = self.get_all_hosts()
        for err_msg in err_msgs:
            logger.warning(gudlog(err_msg, self))
        self.players[config.CONDUCTOR_ALLHOSTS_PLAYER_NAME].hostnames = [
            h.name for h in all_hosts
        ]

        # make johann tarball in prep for tuning
        create_johann_tarball()

        self.state = TaskState.STARTED
        self.started_at = datetime.utcnow()
        # self.started_at = datetime.now(tz=pytz.utc)

        unqueued_measures = copy.copy(self.measures)
        while True:
            self.evaluate_state()

            if self.finished:
                break

            self.smart_queue(unqueued_measures)

            await asyncio.sleep(1)

    def fetch_stored_data(
        self,
        key: str = None,
        subkey: str = None,
        subsubkey: str = None,
        subsubsubkey: str = None,
    ) -> Tuple[bool, Optional[str], int, Any]:
        # variable name level
        if not key:
            return True, None, 200, self.stored_data
        elif key in self.stored_data.keys():
            data = self.stored_data[key]
            if not subkey:
                msg = f"fetched stored data key {key} as {data}"
                return True, msg, 200, data
            # player name level (if not store_singleton)
            elif subkey in data:
                data = data[subkey]

                #  check if iterable
                try:
                    iter(data)
                    iterable = True
                except TypeError:
                    iterable = False

                if not subsubkey:
                    msg = f"fetched stored data key {key}.{subkey} as {data}"
                    return True, msg, 200, data
                elif not iterable:
                    msg = (
                        f"failed to fetch data subkey {subsubkey} from stored data at"
                        f" {key}.{subkey} (data not iterable)"
                    )
                    return False, msg, 400, None
                # host name level (if not store_singleton)
                elif subsubkey in data or (
                    subsubkey.isdigit() and type(data) in [list, dict]
                ):
                    if subsubkey in data:
                        data = data[subsubkey]
                    else:  # fetch dictionary value (host) by index
                        subsubkey = int(subsubkey)
                        try:
                            if type(data) is list:
                                data = data[subsubkey]
                            else:
                                data = list(data.values())[subsubkey]
                        except IndexError:
                            msg = (
                                f"failed to fetch data subkey {subsubkey} from stored"
                                f" data at {key}.{subkey} (index out of bounds)"
                            )
                            return False, msg, 400, None

                    # stored data value level
                    # check if iterable
                    try:
                        iter(data)
                        iterable = True
                    except TypeError:
                        iterable = False

                    if not subsubsubkey:
                        msg = (
                            f"fetched stored data key {key}.{subkey}.{subsubkey} as"
                            f" {data}"
                        )
                        return True, msg, 200, data
                    elif not iterable:
                        msg = (
                            f"failed to fetch data subkey {subsubsubkey} from stored"
                            f" data at {key}.{subkey}.{subsubkey} (data not iterable)"
                        )
                        return False, msg, 400, None
                    elif subsubsubkey in data or (
                        subsubsubkey.isdigit() and type(data) in [list, dict]
                    ):
                        if subsubsubkey in data:
                            data = data[subsubsubkey]
                            msg = (
                                "fetched stored data key"
                                f" {key}.{subkey}.{subsubkey}.{subsubsubkey} as {data}"
                            )
                            return True, msg, 200, data
                        else:
                            subsubsubkey = int(subsubsubkey)
                            try:
                                if type(data) is list:
                                    data = data[subsubsubkey]
                                else:
                                    data = list(data.values())[subsubsubkey]
                            except IndexError:
                                msg = (
                                    f"failed to fetch data subkey {subsubsubkey} from"
                                    f" stored data at {key}.{subkey}.{subsubkey} (index"
                                    " out of bounds)"
                                )
                                return False, msg, 400, None

                            msg = (
                                "fetched stored data key"
                                f" {key}.{subkey}.{subsubkey}.{subsubsubkey} as {data}"
                            )
                            return True, msg, 200, data
                    else:
                        msg = (
                            f"failed to fetch data subkey {subsubsubkey} from stored"
                            f" data at {key}.{subkey}.{subsubkey}"
                        )
                        return False, msg, 400, None
                else:
                    msg = (
                        f"failed to fetch data subkey {subsubkey} from stored data at"
                        f" {key}.{subkey}"
                    )
                    return False, msg, 400, None
            else:
                msg = f"failed to fetch data subkey {subkey} from stored data at {key}"
                return False, msg, 400, None
        else:
            msg = f"failed to fetch stored data key {key}"
            return False, msg, 400, None

    def store_results(
        self, measure: "Measure", player_name: Optional[str], results: Dict
    ) -> None:
        key = measure.store_as
        # initialize the store_as object if we need to
        if key not in self.stored_data:
            self.stored_data[key] = {}

        if player_name:  # i.e. not store_singleton
            if (
                player_name in self.stored_data[key]
                and results != self.stored_data[key][player_name]
            ):
                msg = (
                    f"Overwriting stored data key {key}[{player_name}]. Previous"
                    f" value:\n{self.stored_data[key][player_name]}"
                )
                logger.info(gudlog(msg, self))

            self.stored_data[key][player_name] = results

            msg = f"Stored {key}[{player_name}]:\n{results}"
            logger.debug(gudlog(msg, self))
        else:
            if self.stored_data[key] and results != self.stored_data[key]:
                msg = (
                    f"Overwriting stored data key {key}. Previous"
                    f" value:\n{self.stored_data[key]}"
                )
                logger.info(gudlog(msg, self))

            self.stored_data[key] = results

            msg = f"Stored {key}:\n{results}"
            logger.debug(gudlog(msg, self))

    # this should run in an executor, as it can block an arbitrarily long time
    async def play_the_player(self, measure: "Measure", player: "Player") -> bool:
        delay = measure.start_delay

        # handle special arguments like random numbers and stored values
        try:
            if isinstance(delay, str):
                delay = transform_arg(self, measure, player, delay)
                assert isinstance(delay, int)
            new_args = transform_args(self, measure, player, measure.args)
        except (AssertionError, KeyError, ValueError):
            msg = gudexc(
                "Failed to queue measure -- bad special argument(s)",
                self,
                player,
                measure,
            )
            logger.error(msg)
            measure.state = TaskState.FAILURE
            # we can't set measure.finished yet; other players for this measure may be running
            measure.status["all"]["all"] = f"{msg}; see logs for details"
            return False

        msg = f"queueing measure {measure.name} with a delay of {delay} seconds"
        logger.info(gudlog(msg, self, player))

        msg = f"(transformed) args:\n{pprint.pformat(new_args, indent=4)}"
        logger.debug(gudlog(msg, self, player, measure))

        success, err_msg, group_task = player.enqueue(
            self, measure, measure.task_name, delay, *new_args
        )
        if success:
            measure.celery_group_tasks[player.name] = group_task
        else:
            msg = f"failed to play measure {measure.name}: {err_msg}"
            logger.warning(gudlog(msg, self, player))
            measure.state = TaskState.FAILURE
            measure.status[player.name]["all"] = msg
            return False

    def validate_create_host_mappings(self) -> Tuple[bool, List[str]]:
        success = True
        err_msgs = []

        # map missing (to be created) hostnames to players
        if self.create_hosts:
            self.map_missing_hosts(get_host_names())

        for p in self.players.values():
            # skip local player
            if p.name == config.CONDUCTOR_ALLHOSTS_PLAYER_NAME:
                continue

            # validate hostnames length
            if p.hostnames == [] and not self.create_hosts:
                msg = f"{p.name}: no hosts mapped"
                if p.scale == 0 and config.ALLOW_EMPTY_PLAYER_HOSTS:
                    logger.debug(gudlog(msg, self))
                else:
                    success = False
                    logger.warning(gudlog(msg, self))
                    err_msgs.append(msg)
                    continue
            elif len(p.hostnames) != p.scale and not self.create_hosts:
                success = False
                msg = f"{p.name}: length of hosts does not match scale ({p.scale})"
                logger.warning(gudlog(msg, self))
                err_msgs.append(msg)
                continue

            for host_name in p.hostnames:
                if host_name not in hosts:
                    # create host obj
                    host_dict = {
                        "hostname": host_name,
                        "image": p.image,
                    }
                    try:
                        msg = (
                            f"Temporarily creating Host object for {host_name} with"
                            f" image {p.image}"
                        )
                        logger.debug(gudlog(msg, self, p))
                        host_obj = HostSchema().load(host_dict)
                    except MarshmallowValidationError:
                        success = False
                        msg = f"{host_name}: error creating Host object"
                        logger.warning(gudlog(msg, self, p))
                        err_msgs.append(msg)
                        continue
                elif hosts[host_name].get_image() not in [None, p.image]:
                    success = False
                    msg = (
                        f"{host_name}: Host object already exists with conflicting"
                        " image"
                    )
                    logger.warning(gudlog(msg, self, p))
                    err_msgs.append(msg)
                    continue
                elif hosts[host_name].tuning:
                    success = False
                    msg = (
                        f"host '{host_name}' is tuning or pending tuning; try again"
                        " soon"
                    )
                    logger.warning(gudlog(msg, self, p))
                    err_msgs.append(msg)
                    continue
                else:
                    host_obj = hosts[host_name]

                # make sure we have externally-accessible Redis if needed
                if host_obj.control_method not in config.HOST_CONTROL_EXTERNAL_REDIS:
                    success = False
                    msg = (
                        f"{host_name}: control_method '{host_obj.control_method}'"
                        " is not properly registered in"
                        " config.HOST_CONTROL_EXTERNAL_REDIS"
                    )
                    logger.warning(gudlog(msg, self, p))
                    err_msgs.append(msg)
                    continue
                elif (
                    not config.REDIS_HOST_EXTERNAL
                    and config.HOST_CONTROL_EXTERNAL_REDIS[host_obj.control_method]
                ):
                    success = False
                    msg = (
                        f"host ('{host_name}') requires externally-accessible Redis,"
                        " but environment variable REDIS_HOST_EXTERNAL is not"
                        " specified -- it is usually easiest to include this in the"
                        " conductor's"
                        f" '{config.SRC_ROOT.joinpath(config.ENV_FILE)}'"
                    )
                    logger.warning(gudlog(msg, self, p))
                    err_msgs.append(msg)
                    continue

                # get HostControl object for this Host
                if host_obj.control_method.upper() == "DOCKER":
                    host_control_class = DockerHostControl
                else:
                    host_control_class, msg = get_host_control_class(
                        host_obj.control_method
                    )
                    if not host_control_class:
                        success = False
                        msg = f"{host_name}: {msg}"
                        logger.warning(gudlog(msg, self, p))
                        err_msgs.append(msg)
                        continue

                # check if the host was recently confirmed to be turned on
                host_recently_confirmed_on = False
                if host_obj.last_confirmed_on:
                    check_age = (datetime.utcnow() - host_obj.last_confirmed_on).seconds
                    if check_age < config.HOST_CONFIRMED_ON_VALID_SECS:
                        host_recently_confirmed_on = True
                        msg = (
                            f"Host '{host_name}' with control_name"
                            f" '{host_obj.control_name}' confirmed to be on via"
                            f" {host_obj.control_method} {check_age}s ago; not checking"
                            " again"
                        )
                        logger.debug(msg)

                host_confirmed_on = False
                if not host_recently_confirmed_on:
                    control_name = host_obj.control_name or host_obj.name
                    host_confirmed_on = host_control_class.host_exists(control_name)
                    if host_confirmed_on:
                        host_obj.last_confirmed_on = datetime.utcnow()
                        msg = (
                            f"Host '{host_name}' with control_name"
                            f" '{host_obj.control_name}' appears to exist via"
                            f" {host_obj.control_method}"
                        )
                        logger.debug(msg)

                if host_recently_confirmed_on or host_confirmed_on:
                    if (
                        host_name not in hosts
                    ):  # we may have created host_obj above and not yet in config.hosts
                        hosts[host_name] = host_obj
                        msg = (
                            f"Added new Host object for {host_name} with image"
                            f" {p.image}"
                        )
                        logger.info(gudlog(msg, self, p))
                elif not self.create_hosts:
                    success = False
                    msg = (
                        f"host '{host_name}' with control_name"
                        f" '{host_obj.control_name}' not found via (possibly default)"
                        f" control method '{host_obj.control_method}'"
                    )
                    logger.warning(gudlog(msg, self, p))
                    err_msgs.append(msg)
                    continue
                else:
                    # mark hosts that need to be created
                    hosts[host_name].pending_create = True
                    msg = f"host {host_name} marked for creation"
                    logger.debug(gudlog(msg, self, p))

        return success, err_msgs



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import time
from typing import TYPE_CHECKING, Any, Dict

import celery.exceptions
from marshmallow import EXCLUDE

from johann.host import Host, HostSchema
from johann.host_control_util import get_host_control_class
from johann.shared.config import JohannConfig, celery_app
from johann.shared.logger import JohannLogger
from johann.util import get_codehash

if TYPE_CHECKING:
    from celery import Task

    from johann.host_control import HostControl


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


@celery_app.task(bind=True, autoretry_for=[Exception], retry_backoff=3, max_retries=2)
def tune_host(self: "Task", target_host_dict: Dict[str, Any] = None) -> None:
    score_name = player_name = "_"

    codehash = get_codehash()
    host_copy: "Host" = HostSchema().load(target_host_dict, unknown=EXCLUDE)

    prefix = self.request.shadow
    if host_copy.name == config.CONDUCTOR_LOCAL_HOST_NAME:
        logger.info(f"{prefix}: unnecessary to tune {host_copy.name}")
        return
    elif host_copy.control_method not in config.HOST_CONTROL_CLASS_NAMES:
        msg = f"{prefix}: unrecognized control method '{host_copy.control_method}'"
        self.max_retries = 0
        raise Exception(msg)
    else:
        host_control_class, msg = get_host_control_class(host_copy.control_method)
        if not host_control_class:
            msg = f"{prefix}: {msg}"
            self.max_retries = 0
            raise Exception(msg)
        else:
            host_controller: "HostControl" = host_control_class(host_copy)

        logger.debug(
            f"{prefix}: tuning host '{host_copy.name}', which has"
            f" parameters:\n{host_copy}"
        )
        logger.debug(host_copy)

    # check for Johann on actual host
    do_johann_install = False
    do_johann_update = False
    blank_args = ()
    try:
        sig = host_copy.get_task_signature(
            score_name,
            player_name,
            "tune_orchestra.remote_codehash.pre",
            config.REMOTE_CODEHASH_FUNC,
            0,
            *blank_args,
        )
        remote_codehash = sig.apply_async().get(10, disable_sync_subtasks=False)

        logger.debug(f"{host_copy.name}: received codehash '{remote_codehash}'")
        if remote_codehash != codehash:
            logger.info(f"{host_copy.name}: code hash mismatch")
            do_johann_update = True
    except celery.exceptions.TimeoutError:
        logger.info(f"{host_copy.name}: timed out waiting for remote codehash")
        logger.info(
            f"{host_copy.name}: assuming Johann missing or not running (will perform"
            " install)"
        )
        if config.HOST_AUTO_INSTALL:
            do_johann_install = True

    # install/update latest Johann to player
    if not (do_johann_install or do_johann_update):
        logger.debug(f"{host_copy.name} does not require update or install")
        return

    if host_copy.is_playing():
        raise Exception(
            f"{host_copy.name} is currently playing a score and cannot update Johann"
        )

    success, error_message = host_controller.push_johann(update_only=do_johann_update)

    if not success:
        raise Exception(error_message.strip())

    logger.debug(f"{host_copy.name}: Giving Celery some time to start up")
    time.sleep(15)

    # validate Johann install
    try:
        sig = host_copy.get_task_signature(
            score_name,
            player_name,
            "tune_orchestra.remote_codehash.post",
            config.REMOTE_CODEHASH_FUNC,
            0,
            *blank_args,
        )
        logger.debug(f"{host_copy.name}: requesting post-push codehash")
        remote_codehash = sig.apply_async().get(10, disable_sync_subtasks=False)
        logger.debug(
            f"{host_copy.name}: received post-push codehash '{remote_codehash}'"
        )

        if remote_codehash != codehash:
            logger.debug(f"{codehash} vs {remote_codehash}")
            msg = f"{host_copy.name}: push failed; code hash mismatch"
            logger.warning(msg)
            raise Exception(msg)
    except celery.exceptions.TimeoutError:
        msg = f"{host_copy.name}: push failed; timed out waiting for remote codehash"
        logger.warning(msg)
        raise Exception(msg)

    logger.info(f"{host_copy.name} successfully tuned")
    return



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import copy
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple
from uuid import UUID, uuid4

from celery.canvas import signature
from marshmallow import Schema
from marshmallow import ValidationError as MarshmallowValidationError
from marshmallow import fields, post_load, validates
from marshmallow_enum import EnumField

from johann.shared.config import JohannConfig, celery_app
from johann.shared.enums import HostOS, PmtrVariant, TaskState
from johann.shared.fields import LaxStringField, NameField
from johann.shared.logger import JohannLogger
from johann.util import gudlog, safe_name

if TYPE_CHECKING:
    from celery.canvas import Signature


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


class HostSchema(Schema):
    class Meta:
        ordered = True

    name = NameField(
        required=True, data_key="hostname"
    )  # hostname for network use (i.e. fqdn, hostname, or, if necessary, IP)
    control_name = fields.Str(
        allow_none=True, missing=None
    )  # i.e., container ID for DOCKER (optional)
    johann_id = fields.UUID(allow_none=True, missing=None)
    image = fields.Str(allow_none=True, missing=None)
    user = fields.Str(allow_none=True, missing=None)
    pwd_env = fields.Str(allow_none=True, missing=None)
    os = EnumField(HostOS, allow_none=True, missing=None)
    python_path = fields.Str(allow_none=True, missing=None)
    python_ver = LaxStringField(allow_none=True, missing=None)
    pmtr_variant = EnumField(PmtrVariant, allow_none=True, missing=None)
    control_method = fields.Str(allow_none=True, missing=None)
    pip_offline_install = fields.Boolean(allow_none=True, missing=None)

    tuning = fields.Boolean(dump_only=True)
    pending_create = fields.Boolean(dump_only=True)
    celery_task_ids = fields.List(
        fields.Str(), dump_only=True
    )  # note that finished tasks may be cleared from this list at any time
    last_checked_exists = fields.DateTime(dump_only=True)

    @validates("python_ver")
    def validate_python_ver(self, value):
        if value is not None and value not in config.SUPPORTED_PYTHON_VERSIONS:
            raise MarshmallowValidationError(f"Unsupported python version '{value}'")

    @validates("control_method")
    def validate_control_method(self, value):
        if value is not None and value not in config.HOST_CONTROL_CLASS_NAMES:
            raise MarshmallowValidationError(f"Unrecognized control method '{value}'")

    @post_load
    def make_host(self, data: Dict[str, Any], **kwargs) -> "Host":
        name = data["name"]
        control_name = data["control_name"]

        # validate name and control_name
        if name != safe_name(name):
            raise MarshmallowValidationError(
                f"Name '{name}' does not appear to be a valid hostname"
            )
        if control_name and control_name != safe_name(control_name):
            raise MarshmallowValidationError(
                f"Control name '{control_name}' does not appear to be valid"
            )

        # validate pmtr variant
        if data["pmtr_variant"]:
            try:
                PmtrVariant(data["pmtr_variant"])
            except ValueError:
                raise MarshmallowValidationError(
                    f"PMTR variant '{data['pmtr_variant']}' is invalid"
                )

        # validate control method
        if (
            data["control_method"]
            and data["control_method"] not in config.HOST_CONTROL_CLASS_NAMES
        ):
            raise MarshmallowValidationError(
                f"Control method '{data['control_method']}' is invalid"
            )

        return Host(**data)


class Host(object):
    def __init__(
        self,
        name,
        control_name,
        johann_id,
        image,
        user,
        pwd_env,
        os,
        python_path,
        python_ver,
        pmtr_variant,
        control_method,
        pip_offline_install,
    ) -> None:
        self.name: str = name
        self.control_name: Optional[str] = control_name
        self.johann_id: Optional[UUID] = johann_id
        self._image: Optional[str] = image
        self.user: Optional[str] = user
        self.pwd_env: Optional[str] = pwd_env
        self.os: Optional[HostOS] = None  # initialized below
        self.python_path: Optional[str] = python_path
        self.python_ver: Optional[str] = python_ver
        self.pmtr_variant: Optional[PmtrVariant] = None  # initialized below
        self.control_method: Optional[str] = None  # initialized below
        self.pip_offline_install: bool = pip_offline_install

        self.tuning: bool = False
        self.pending_create: bool = False
        self.celery_task_ids: List[
            str
        ] = []  # note that finished tasks may be cleared from this list at any time
        # only tasks that are currently running can be relied upon to be here
        self.last_confirmed_on: Optional[datetime] = None

        logger.debug(f"Creating Host object for '{name}'")

        if pmtr_variant:
            try:
                self.pmtr_variant = PmtrVariant(pmtr_variant)
            except ValueError:
                msg = f"{pmtr_variant} is not a valid PMTR variant; ignoring..."
                logger.warning(msg)
                self.pmtr_variant = None  # will try to match to image below

        if os:
            try:
                self.os = HostOS(os)
            except ValueError:
                msg = (
                    f"{os} is not a valid OS option for Johann at this time;"
                    " ignoring..."
                )
                logger.warning(msg)
                self.os = None  # will try to match to image below

        if (
            not self.user
            and self._image in config.HOST_IMAGE_PARAMS
            and "user" in config.HOST_IMAGE_PARAMS[self._image]
        ):
            self.match_user_to_image()
        if not self.pwd_env:
            self.match_pwd_env_to_image()
        if (
            not self.os
            and self._image in config.HOST_IMAGE_PARAMS
            and "os" in config.HOST_IMAGE_PARAMS[self._image]
        ):
            self.match_os_to_image()
        if not self.python_path:
            self.match_python_path_to_image()
        if not self.python_ver:
            self.match_python_ver_to_image()
        if not self.pmtr_variant:
            self.match_pmtr_variant_to_image()
        if not self.control_method:
            self.match_control_method_to_image()
        if not self.pip_offline_install:
            self.match_pip_offline_install_to_image()

    def get_image(self) -> Optional[str]:
        return self._image

    def set_image(self, image: Optional[str]) -> bool:
        if self._image == image:
            return True

        self._image = image
        ret = True

        ret &= self.match_user_to_image()
        ret &= self.match_pwd_env_to_image()
        ret &= self.match_os_to_image()
        ret &= self.match_python_path_to_image()
        ret &= self.match_pmtr_variant_to_image()
        ret &= self.match_control_method_to_image()
        ret &= self.match_pip_offline_install_to_image()

        return ret

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(hostname={self.name},"
            f"control_name={self.control_name},image={self._image},os={self.os},"
            f"control={self.control_method})"
        )

    def dump(self) -> Dict[str, Any]:
        selfcopy = copy.deepcopy(self)
        selfcopy.image = self._image
        return HostSchema().dump(selfcopy)

    def dumps(self) -> str:
        selfcopy = copy.deepcopy(self)
        selfcopy.image = self._image
        return HostSchema().dumps(selfcopy)

    def copy_from(self, host: "Host") -> Tuple[bool, Optional[str]]:
        if not isinstance(host, Host):
            return False, "not a valid Host object"

        if host.name != self.name:
            logger.warning(
                "copy_from() called with a host with a different name...was this"
                " intentional?"
            )

        self.name = host.name

        if host.get_image() != self._image:
            msg = f"updating image from {self._image} to {host.get_image()}"
            logger.debug(msg)
            self.set_image(host.get_image())

        if host.user != self.user:
            msg = f"updating user from {self.user} to {host.user}"
            logger.debug(msg)
            self.user = host.user

        if host.pwd_env != self.pwd_env:
            msg = f"updating pwd_env from {self.pwd_env} to {host.pwd_env}"
            logger.debug(msg)
            self.pwd_env = host.pwd_env

        if host.os != self.os:
            msg = f"updating os from {self.os} to {host.os}"
            logger.debug(msg)
            self.os = host.os

        if host.python_path != self.python_path:
            msg = f"updating python path from {self.python_path} to {host.python_path}"
            logger.debug(msg)
            self.python_path = host.python_path

        if host.python_ver != self.python_ver:
            msg = f"updating python version from {self.python_ver} to {host.python_ver}"
            logger.debug(msg)
            self.python_ver = host.python_ver

        if host.pmtr_variant != self.pmtr_variant:
            msg = (
                f"updating pmtr variant from {self.pmtr_variant.value} to"
                f" {host.pmtr_variant.value}"
            )
            logger.debug(msg)
            self.pmtr_variant = host.pmtr_variant

        if host.control_method != self.control_method:
            msg = (
                f"updating control method from {self.control_method} to"
                f" {host.control_method}"
            )
            logger.debug(msg)
            self.control_method = host.control_method

        if host.pip_offline_install != self.pip_offline_install:
            msg = (
                f"updating pip offline install from {self.pip_offline_install} to"
                f" {host.pip_offline_install}"
            )
            logger.debug(msg)
            self.pip_offline_install = host.pip_offline_install

        return True, None

    def match_user_to_image(self) -> bool:
        try:
            self.user = config.HOST_IMAGE_PARAMS[self._image]["user"]
            logger.debug(f"matched user '{self.user}' to image '{self._image}'")
            return True
        except KeyError:
            logger.info(
                f"{self.name}: unable to find matching user for image '{self._image}'"
                " in config; using default"
            )
            self.user = config.DEFAULT_PLAYER_USER
            return False

    def match_pwd_env_to_image(self) -> bool:
        try:
            self.pwd_env = config.HOST_IMAGE_PARAMS[self._image]["pwd_env"]
            logger.debug(f"matched pwd_env '{self.pwd_env}' to image '{self._image}'")
            return True
        except KeyError:
            logger.debug(
                f"{self.name}: unable to find matching pwd_env for image"
                f" '{self._image}' in config; using default"
            )
            self.pwd_env = config.DEFAULT_PLAYER_PWD_ENV
            return False

    def match_os_to_image(self) -> bool:
        try:
            image_value = config.HOST_IMAGE_PARAMS[self._image]["os"]
            try:
                self.os = HostOS(image_value)
                logger.debug(f"matched os '{self.os}' to image '{self._image}'")
                return True
            except ValueError:
                msg = (
                    f"{image_value} is not a valid OS option at this time; using"
                    " default"
                )
                logger.info(msg)
                self.os = HostOS(config.DEFAULT_PLAYER_OS)
                return False
        except KeyError:
            logger.info(
                f"{self.name}: unable to find matching os for image '{self._image}' in"
                " config; using default"
            )
            self.os = config.DEFAULT_PLAYER_OS
            return False

    def match_python_path_to_image(self) -> bool:
        try:
            self.python_path = config.HOST_IMAGE_PARAMS[self._image]["python_path"]
            logger.debug(
                f"matched python '{self.python_path}' to image '{self._image}'"
            )
            return True
        except KeyError:
            logger.info(
                f"{self.name}: unable to find matching python path for image"
                f" '{self._image}' in config; using default"
            )
            self.python_path = config.DEFAULT_PYTHON_PATH
            return False

    def match_python_ver_to_image(self) -> bool:
        try:
            self.python_ver = config.HOST_IMAGE_PARAMS[self._image]["python_ver"]
            logger.debug(
                f"matched python version '{self.python_ver}' to image '{self._image}'"
            )
            return True
        except KeyError:
            logger.debug(
                f"{self.name}: unable to find matching python version for image"
                f" '{self._image}' in config; will be discovered at runtime"
            )
            self.python_ver = None
            return False

    def match_pmtr_variant_to_image(self) -> bool:
        try:
            image_value = config.HOST_IMAGE_PARAMS[self._image]["pmtr_variant"]
            try:
                self.pmtr_variant = PmtrVariant(image_value)
                logger.debug(
                    f"matched pmtr variant '{self.pmtr_variant.value}' to image"
                    f" '{self._image}'"
                )
                return True
            except ValueError:
                msg = f"{image_value} is not a valid PMTR variant; using default"
                logger.warning(msg)
                self.pmtr_variant = PmtrVariant(config.DEFAULT_PMTR_VARIANT)
                return False
        except KeyError:
            logger.info(
                f"{self.name}: unable to find matching pmtr variant for image"
                f" '{self._image}' in config; using default"
            )
            self.pmtr_variant = PmtrVariant(config.DEFAULT_PMTR_VARIANT)
            return False

    def match_control_method_to_image(self) -> bool:
        try:
            image_value = config.HOST_IMAGE_PARAMS[self._image]["control_method"]
            if image_value not in config.HOST_CONTROL_CLASS_NAMES:
                msg = f"{image_value} is not a valid control method; using default"
                logger.warning(msg)
                self.control_method = config.DEFAULT_HOST_CONTROL
                return False
            else:
                self.control_method = image_value
                logger.debug(
                    f"matched control method '{self.control_method}' to image"
                    f" '{self._image}'"
                )
                return True
        except KeyError:
            logger.debug(
                f"{self.name}: unable to find matching control method for image"
                f" '{self._image}' in config; using default"
            )
            self.control_method = config.DEFAULT_HOST_CONTROL
            return False

    def match_pip_offline_install_to_image(self) -> bool:
        try:
            self.pip_offline_install = config.HOST_IMAGE_PARAMS[self._image][
                "pip_offline_install"
            ]
            logger.debug(
                f"matched pip offline install '{self.pip_offline_install}' to image"
                f" '{self._image}'"
            )
            return True
        except KeyError:
            logger.debug(
                f"{self.name}: unable to find matching pip offline install for image"
                f" '{self._image}' in config; using default"
            )
            self.pip_offline_install = config.DEFAULT_PIP_OFFLINE_INSTALL
            return False

    def is_playing(self) -> bool:
        self.clear_finished_celery_task_ids()
        if len(self.celery_task_ids) > 0:
            return True
        else:
            return False

    def clear_finished_celery_task_ids(self) -> None:
        to_remove = []
        for cti in self.celery_task_ids:
            ct = celery_app.AsyncResult(cti)
            if ct.state in [TaskState.SUCCESS, TaskState.FAILURE]:  # final states
                to_remove.append(ct)

        for cti in to_remove:
            self.celery_task_ids.remove(cti)

    def get_task_signature(
        self,
        score_name: str,
        player_name: str,
        measure_name: str,
        func: str,
        delay: int,
        *task_args,
        **task_kwargs,
    ) -> "Signature":
        if self.tuning:
            msg = (
                "host is tuning or pending tuning; strongly advise against running new"
                " tasks on it or they may be interrupted without warning or recovery"
            )
            logger.warning(gudlog(msg, score_name, player_name, None, self.name))
        description = f"{score_name}.{player_name}.{self.name}.{measure_name}"
        task_id = str(
            uuid4()
        )  # we need to know the task_id a priori for score.task_map
        sig_opts = {
            "queue": self.name,
            "shadow": description,
            "task_id": task_id,
            "countdown": delay,
        }
        sig = signature(func, args=task_args, kwargs=task_kwargs, options=sig_opts)
        msg = (
            f"task signature created for {func}{task_args} with"
            f" kwargs:\n{task_kwargs}\nand options:\n{sig_opts}"
        )
        logger.log(5, gudlog(msg, score_name, player_name, measure_name, self.name))

        return sig



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import re
from typing import Any, Optional

import marshmallow

from johann.shared.enums import TaskState


class NameField(marshmallow.fields.String):
    """A name field."""

    default_error_messages = {
        "invalid_name": (
            "Names may only consist of letters, numbers, underscores, and hyphens"
        )
    }

    def _validated(self, value: Any) -> Optional[str]:
        if value is None:
            return None
        if isinstance(value, str) and re.match(r"[\w\-]+", value):
            return value
        else:
            raise self.make_error("invalid_name")

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[str]:
        return self._validated(value)


class LaxStringField(marshmallow.fields.String):
    """A string field that will attempt to cast non-strings."""

    default_error_messages = {"invalid_string": "neither a string nor castable to one"}

    def _validated(self, value: Any) -> Optional[str]:
        if value is None:
            return None
        if isinstance(value, str):
            return value
        try:
            return str(value)
        except (ValueError, AttributeError, TypeError) as e:
            raise self.make_error("invalid_string") from e

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[str]:
        return self._validated(value)


class StateField(marshmallow.fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        if type(value) is str:
            return value
        else:
            return value.value

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return TaskState(value)
        except ValueError:
            raise marshmallow.ValidationError("not a valid TaskState")



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import os
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import docker
from docker.errors import DockerException
from johann.host_control import HostControl
from johann.shared.config import JohannConfig
from johann.shared.logger import JohannLogger

if TYPE_CHECKING:
    from docker.models.containers import Container
    from johann.host import Host
    from johann.util import PathLikeObj


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger

api_client = docker.APIClient()
client = docker.from_env()


class DockerHostControl(HostControl):
    def __init__(self, host_copy: "Host"):
        super().__init__(host_copy)
        attempted_name = self.control_name or self.name
        self.container = self.get_container_from_name(attempted_name)
        if not self.container:
            raise Exception(
                f"{self.name}: Unable to find container for '{attempted_name}'"
            )

    @staticmethod
    def get_host_names() -> List[str]:
        ret = []
        for c in api_client.containers():
            for n in c["Names"]:
                ret.append(n)
        return ret

    @staticmethod
    def host_exists(name: str) -> bool:
        for c in api_client.containers():
            for n in c["Names"]:
                if name == n.replace("/", "", 1):
                    return True
        return False

    @staticmethod
    def get_container_from_name(name: str) -> Optional["Container"]:
        for c in api_client.containers():
            for n in c["Names"]:
                if name == n.replace("/", "", 1):
                    return client.containers.get(c["Id"])

        return None

    def put_archive(
        self,
        archive_path: "PathLikeObj",
        dest_path_inc_filename: "PathLikeObj",
        remove_archive_file: bool = True,
    ) -> bool:
        with open(archive_path, "rb") as f:
            archive_bytes = f.read()
            return self.container.put_archive(
                str(os.path.dirname(dest_path_inc_filename)), archive_bytes
            )

    def run_cmd(
        self,
        cmd: str,
        environment: Optional[Dict[str, str]] = None,
        detach: bool = False,
        privileged: bool = False,
        workpath: "PathLikeObj" = None,
        finish_timeout: int = -1,
        strip_output: bool = True,
    ) -> Tuple[Optional[int], Optional[str]]:
        # docker exec_run does not support a timeout
        if finish_timeout != -1 and not detach:
            logger.warning(
                f"{self.name}: run_cmd: docker exec_run does not support a timeout;"
                " ignoring..."
            )

        logger.debug(
            f"{self.name}: {cmd} (detach={detach}, privileged={privileged},"
            f" workpath={workpath}), environment={environment}"
        )

        kwargs = {"detach": detach, "privileged": privileged}
        if workpath:
            kwargs["workdir"] = str(workpath)
        if self.user:
            kwargs["user"] = self.user
        if environment:
            kwargs["environment"] = environment

        try:
            r = self.container.exec_run(cmd, **kwargs)

            # r.output, against documentation specs I might add, is type str when detach=True
            output = r.output.decode() if isinstance(r.output, bytes) else r.output

            if detach:
                return None, None
            else:
                if strip_output:
                    output = output.strip()
                return r.exit_code, output
        except DockerException:
            raise



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import json
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from marshmallow import Schema
from marshmallow import ValidationError as MarshmallowValidationError
from marshmallow import fields, post_load, validates

from johann.shared.config import JohannConfig
from johann.shared.enums import TaskState
from johann.shared.fields import NameField, StateField
from johann.shared.logger import JohannLogger
from johann.util import (
    celery_group_status,
    get_attr,
    gudlog,
    parse_special_arg,
    task_state_priority,
)

if TYPE_CHECKING:
    from celery.result import GroupResult

    from johann.score import Score


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


class MeasureSchema(Schema):
    class Meta:
        ordered = True

    name = NameField(required=True)
    player_names = fields.List(fields.Str(), required=True, data_key="players")
    task_name = fields.Str(required=True, data_key="task")
    args = fields.List(fields.Raw(allow_none=True), missing=list)
    store_as = fields.Str(allow_none=True, missing=None)
    store_singleton = fields.Boolean(missing=False)
    store_interim_results = fields.Boolean(missing=False)
    lazy_fetch_stored = fields.Boolean(missing=False)

    start_delay = fields.Raw(allow_none=True, missing=None)  # must be integer or string
    depends_on = fields.List(fields.Str(), missing=list)
    dependency_proof = fields.Boolean(missing=False)

    # do not remove any of these three without careful consideration
    state = StateField(dump_only=True)
    status = fields.Dict(
        keys=fields.Str(),  # player
        values=fields.Dict(
            keys=fields.Str(),
            values=fields.Str(),  # host, host status
        ),
        dump_only=True,
    )
    finished = fields.Boolean(dump_only=True)

    celery_group_tasks = fields.Method(
        "get_celery_group_tasks_serializable", dump_only=True
    )
    local_measure = fields.Boolean(dump_only=True)

    @validates("task_name")
    def validate_task_name(self, value):
        attr, msg = get_attr(value)
        if not attr:
            raise MarshmallowValidationError(msg)

    @validates("start_delay")
    def validate_start_delay(self, value):
        msg = "start_delay must be an integer or a special argument string"
        if isinstance(value, str):
            try:
                parse_special_arg(value)
            except ValueError:
                raise MarshmallowValidationError(msg)
        elif value is not None and not isinstance(value, int):
            raise MarshmallowValidationError(msg)

    def get_celery_group_tasks_serializable(self, obj: "Measure"):
        return obj.get_task_status(
            short=True
        )  # short excludes task results, which may not be serializable

    def validate_start_params(self, measure: "Measure"):
        if measure.start_delay is None and measure.depends_on == []:
            raise MarshmallowValidationError(
                "Input data must specify one or both of 'start_delay' and 'depends_on'"
            )

    def validate_store_params(self, measure: "Measure"):
        if measure.store_interim_results and not measure.store_as:
            raise MarshmallowValidationError(
                "'store_as' must be specified if 'store_interim_results' is True"
            )
        if measure.store_singleton:
            if len(measure.player_names) > 1:
                logger.warning(
                    "'store_singleton' with more than one player will likely cause data"
                    " loss"
                )
            if not measure.store_as:
                raise MarshmallowValidationError(
                    "'store_as' must be specified if 'store_singleton' is True"
                )

    def cant_depend_on_yourself(self, measure: "Measure"):
        for depname in measure.depends_on:
            if depname == measure.name:
                raise MarshmallowValidationError("You cannot depend on yourself")

    @post_load(pass_original=True)
    def make_measure(
        self, data: Dict[str, Any], original_data: Dict[str, Any], **kwargs: Any
    ) -> "Measure":
        measure = Measure(**data, original_data=original_data)
        self.validate_start_params(measure)
        self.validate_store_params(measure)
        self.cant_depend_on_yourself(measure)
        return measure


class Measure(object):
    def __init__(
        self,
        name,
        player_names,
        task_name,
        args,
        store_as,
        store_singleton,
        store_interim_results,
        lazy_fetch_stored,
        start_delay,
        depends_on,
        dependency_proof,
        original_data=None,
    ) -> None:
        self.name: str = name
        self.player_names: List[str] = player_names
        self.task_name: str = task_name
        self.args: List = args
        self.store_as: str = store_as
        self.store_singleton: bool = store_singleton
        self.store_interim_results: bool = store_interim_results
        self.lazy_fetch_stored: bool = lazy_fetch_stored
        self.start_delay: int = start_delay
        self.depends_on: List[str] = depends_on
        self.dependency_proof: bool = dependency_proof
        self.original_data: Dict[str, Any] = original_data

        self.state: TaskState = TaskState.PENDING
        self.status: Dict[str, Dict[str, str]] = defaultdict(dict)
        self.finished: bool = False
        self.celery_group_tasks: Dict[str, "GroupResult"] = {}  # keys are Player names
        self.local_measure: bool = False  # whether or not run locally by conductor

    def __repr__(self) -> str:
        format_str = (
            f"{self.__class__.__name__}(name={self.name},"
            f"player_names={'|'.join(self.player_names)},task_name={self.task_name},"
            f"start_delay={self.start_delay},depends_on={'|'.join(self.depends_on)},"
            f"dependency_proof={self.dependency_proof})"
        )
        return format_str

    def dump(self) -> Dict[str, Any]:
        return MeasureSchema().dump(self)

    def dumps(self) -> str:
        return MeasureSchema().dumps(self)

    def store_results(
        self, score: "Score", task_status: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        if not self.store_as:
            return None

        results = {}
        for player_name, tstat in task_status.items():
            results_p = {}
            for t_id, t in tstat["tasks"].items():
                t_result = None

                if "result" in t:
                    t_result = t["result"]
                elif (
                    self.store_interim_results
                    and "meta" in t
                    and t["meta"]
                    and "interim_result" in t["meta"]
                ):
                    t_result = t["meta"]["interim_result"]

                if t_result is not None:
                    if self.store_singleton:
                        score.store_results(self, None, t_result)

                    if t_id not in score.task_map:
                        msg = (
                            f"task {t_id} not found in task_map. This is probably bad."
                        )
                        logger.info(gudlog(msg, score, player_name, self))
                        results_p[t_id] = t_result
                    else:
                        results_p[score.task_map[t_id]["host_name"]] = t_result

            results[player_name] = results_p

            if results_p and not self.store_singleton:
                score.store_results(self, player_name, results_p)

        return results

    def evaluate_state(self, score: "Score") -> None:
        task_status = self.get_task_status(short=False)

        prior_state = self.state

        success = 0
        finished = 0
        for player_name, tstat in task_status.items():
            if tstat["finished"]:
                finished += 1
            if tstat["state"] == TaskState.SUCCESS:
                success += 1

            if task_state_priority(tstat["state"]) > task_state_priority(self.state):
                self.state = tstat["state"]

            if tstat["status"]:
                self.status[player_name] = tstat["status"]

        if self.state == TaskState.FAILURE and finished == len(task_status):
            self.finished = True

        if success == len(self.player_names):
            self.state = TaskState.SUCCESS
        if finished == len(self.player_names):
            self.finished = True

        if self.state == TaskState.FAILURE and self.state != prior_state:
            msg = f"measure '{self.name}' failed:\n{json.dumps(self.status, indent=2)}"
            logger.warning(gudlog(msg, score))

        self.store_results(score, task_status)

    def started(self) -> bool:
        if self.state not in [TaskState.PENDING]:
            return True
        else:
            return False

    def get_status(self, short: bool = False) -> Dict[str, Any]:
        status = {}
        if self.start_delay is not None:
            status["start_delay"] = self.start_delay
        if self.depends_on is not []:
            status["depends_on"] = self.depends_on
        status["dependency_proof"] = self.dependency_proof
        status["state"] = self.state
        if self.status or not short:
            status["status"] = self.status
        status["finished"] = self.finished
        status["task_status"] = self.get_task_status(short=short)
        status["local_measure"] = self.local_measure

        return status

    def get_task_status(self, short: bool = False) -> Dict[str, Any]:
        status = {}

        for player_name, group_task in self.celery_group_tasks.items():
            status[player_name] = celery_group_status(group_task, short=short)

        return status



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
from johann.shared.config import JohannConfig

config = JohannConfig.get_config()
broker_url = config.REDIS_URL
result_backend = config.REDIS_URL
task_track_started = True



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple
from uuid import uuid4

from celery import group, signature
from marshmallow import Schema, fields, post_load

from johann.shared.config import JohannConfig, hosts
from johann.shared.fields import NameField
from johann.shared.logger import JohannLogger
from johann.util import gudlog

if TYPE_CHECKING:
    from celery.canvas import Signature
    from celery.result import GroupResult

    from johann.host import Host
    from johann.measure import Measure
    from johann.score import Score


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


class PlayerSchema(Schema):
    class Meta:
        ordered = True

    name = NameField(required=True)
    image = fields.Str(allow_none=True, missing=None)
    hostnames = fields.List(
        fields.Str(), data_key=config.PLAYER_HOSTS_DUMP_KEY, missing=list
    )
    scale = fields.Integer(missing=-1)

    @post_load
    def make_player(self, data: Dict[str, Any], **kwargs: Any) -> "Player":
        if data["scale"] < 0:
            data["scale"] = max(1, len(data["hostnames"]))
        return Player(**data)


class Player(object):
    def __init__(self, name, image, hostnames, scale) -> None:
        self.name: str = name
        self.image: str = image
        self.hostnames: List[str] = hostnames
        self.scale: int = scale

    def __repr__(self) -> str:
        return "{}(name={},image={},hostnames={},scale={})".format(
            self.__class__.__name__,
            self.name,
            self.image,
            "|".join(self.hostnames),
            self.scale,
        )

    def dump(self) -> Dict[str, Any]:
        return PlayerSchema().dump(self)

    def dumps(self) -> str:
        return PlayerSchema().dumps(self)

    def copy_from(
        self, player: "Player", score: "Score"
    ) -> Tuple[Optional[bool], Optional[str]]:
        """

        Args:
            player:
            score:

        Returns:
            None if no changes
            True if changed successfully
            False if player is not a Player object

        """
        if not isinstance(player, Player):
            return False, "not a Player object"

        if player.name != self.name:
            logger.warning(
                "copy_from() called with a player with a different name...was this"
                " intentional?"
            )

        self.name = player.name
        changed = False

        if player.hostnames != self.hostnames:
            changed = True
            msg = "updating hostnames from {} to {}".format(
                self.hostnames, player.hostnames
            )
            logger.debug(gudlog(msg, score, self))
            self.hostnames = player.hostnames

        if player.scale != self.scale:
            changed = True
            msg = "updating scale from {} to {}".format(self.scale, player.scale)
            logger.debug(gudlog(msg, score, self))
            self.scale = player.scale

        if player.image != self.image:
            changed = True
            msg = "updating image from {} to {}".format(self.image, player.image)
            logger.debug(gudlog(msg, score, self))
            self.image = player.image

        if not changed:
            return None, None
        else:
            return True, None

    @staticmethod
    def get_local_task_signature(
        score_name: str,
        measure_name: str,
        host: "Host",
        func: str,
        delay: int,
        *args: Any,
    ) -> "Signature":
        description = f"{score_name}.LOCAL.{measure_name}.{host.name}"
        task_id = str(
            uuid4()
        )  # we need to know the task_id a priori for score.task_map
        sig_opts = {
            "queue": config.CELERY_QUEUE_ID,
            "shadow": description,
            "task_id": task_id,
            "countdown": delay,
        }

        sig = signature(
            func, args=args, kwargs={"target_host_dict": host.dump()}, options=sig_opts
        )
        msg = "task signature created for {}{} with options:\n{}".format(
            func, args, sig_opts
        )
        logger.log(5, f"{description}| {msg}")

        return sig

    def enqueue(
        self, score: "Score", measure: "Measure", func: str, delay: int, *args: Any
    ) -> Tuple[bool, Optional[str], Optional["GroupResult"]]:
        signatures = []

        for hostname in self.hostnames:
            if hostname not in hosts:
                msg = "{} not found in dictionary of hosts".format(hostname)
                logger.warning(gudlog(msg, score, self, measure.name))
                return False, msg, None
            host = hosts[hostname]

            if measure.local_measure:
                sig = Player.get_local_task_signature(
                    score.name, measure.name, host, func, delay, *args
                )
            else:
                if host.pending_create:
                    msg = "{} still pending creation".format(host.name)
                    logger.warning(gudlog(msg, score, self, measure.name))
                    return False, msg, None

                sig = host.get_task_signature(
                    score.name, self.name, measure.name, func, delay, *args
                )
                host.clear_finished_celery_task_ids()
                host.celery_task_ids.append(sig.id)

            if sig is None:
                msg = "task signature creation failed for hostname {}".format(host.name)
                logger.warning(gudlog(msg, score, self, measure.name))
                return False, msg, None

            score.task_map[sig.id] = {
                "measure_name": measure.name,
                "player_name": self.name,
                "host_name": host.name,
            }
            signatures.append(sig)

        task_group = group(signatures)

        group_result = task_group.apply_async()

        return True, None, group_result



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import glob
import hashlib
import importlib
import json
import os.path
import pkgutil
import random
import re
import subprocess
import sys
from pathlib import PurePath
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import aiohttp.web
import pkg_resources

from johann.shared.config import JohannConfig, active_plugins, celery_app
from johann.shared.enums import TaskState
from johann.shared.logger import JohannLogger

if TYPE_CHECKING:
    from asyncio import Future
    from typing import TypeVar

    from celery.result import AsyncResult, GroupResult

    from johann.host import Host
    from johann.measure import Measure
    from johann.player import Player
    from johann.score import Score

    PathLikeObj = TypeVar("PathLikeObj", str, bytes, PurePath)


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


def task_state_priority(state: TaskState) -> int:
    try:
        state = TaskState(state)
    except ValueError:
        msg = f"{state} is not a valid TaskState"
        logger.error(msg)
        raise

    sorted_pri = [  # low to high
        TaskState.SUCCESS,
        TaskState.PENDING,
        TaskState.DEFERRED,
        TaskState.QUEUED,
        TaskState.STARTED,
        TaskState.PROGRESS,
        TaskState.RETRY,
        TaskState.FAILURE,
    ]

    try:
        return sorted_pri.index(state)
    except ValueError:
        msg = f"{state} is not a valid TaskState"
        logger.exception(msg)
        raise


def get_codehash() -> str:
    if not config.CODEHASH:
        config.CODEHASH = calculate_codehash()
        logger.debug(f"Codehash: {config.CODEHASH}")

    return config.CODEHASH


def _validate_score_name_dir(
    package_name: str, score_name: str, score_dir: str = "scores"
) -> bool:

    if safe_name(score_dir) != score_dir:
        logger.debug(f"Invalid directory '{score_dir}'")
        return False
    elif safe_name(score_name) != score_name:
        logger.debug(f"Invalid score name '{score_name}'")
        return False
    elif not pkg_resources.resource_isdir(package_name, score_dir):
        logger.debug(f"Directory {score_dir} not found in package {package_name}")
        return False
    else:
        return True


def get_score_str(
    score_name: str,
    package_name: str = None,
    score_dir: str = "scores",
) -> Optional[str]:
    if package_name:
        return _get_score_str_helper(score_name, package_name, score_dir)

    for package_name in ["johann"] + active_plugins:
        score_str = _get_score_str_helper(
            score_name, package_name, score_dir, suppress_warnings=True
        )
        if score_str:
            logger.debug(f"Found score {score_name} in package {package_name}")
            return score_str

    return None


def _get_score_str_helper(
    score_name: str,
    package_name: str,
    score_dir: str = "scores",
    suppress_warnings: bool = False,
) -> Optional[str]:
    """Get a package's score file as a string.

    Args:
        score_name:
            The name of the score.
        package_name:
            The name of the Python package.
        score_dir:
            Optional; The package subdirectory in which to look. Defaults to "scores".
        suppress_warnings:
            Optional; Dont log warnings. Defaults to False.

    Returns:
        The score as a string, or None.
    """
    # validate score_name and score_dir
    if not _validate_score_name_dir(package_name, score_name, score_dir):
        return None

    # try to read score with .yml and .yaml extensions
    try:
        score_path = f"{score_dir}/{score_name}.yml"
        return pkg_resources.resource_string(package_name, score_path)
    except FileNotFoundError:
        try:
            score_path = f"{score_dir}/{score_name}.yaml"
            return pkg_resources.resource_string(package_name, score_path)
        except FileNotFoundError:
            logger.debug(f"'{score_name}' not found in '{score_dir}' of {package_name}")
    except Exception:
        if not suppress_warnings:
            logger.warning(
                f"Error fetching {score_name} from {package_name}", exc_info=True
            )
        return None


def calculate_codehash() -> str:
    logger.debug("Getting codehash...")
    hashes = {}
    for pathname in config.CODEHASH_FILES:
        filenames = glob.glob(pathname)
        for filename in filenames:
            try:
                with open(filename, "rb") as f:
                    filehash = hashlib.sha256()
                    filehash.update(f.read())
                    filehash = filehash.hexdigest()
                    logger.debug(f"{filehash} {filename}")
                    hashes[filename] = filehash
            except OSError as e:
                logger.warning(
                    f"Error while getting codehash ('{filename}'): {e.strerror}"
                )
                continue
    codehash = hashlib.sha256()
    codehash.update(json.dumps(sorted(list(hashes.values()))).encode("utf-8"))
    codehash = codehash.hexdigest()
    logger.debug(f"Codehash: {codehash}")
    return codehash


async def wrap_future(
    fut: "Future",
    future_name: str,
    callback: Optional[Callable],
    *callback_args: Any,
    score_or_measure: Union["Score", "Measure", None] = None,
) -> Any:
    if not future_name:
        future_name = "?"
    try:
        result = await fut
        if callback:
            try:
                callback_result = await callback(result, *callback_args)
                logger.debug(f"{future_name}| callback result: {callback_result}")
            except Exception as e:
                msg = f"{future_name}| callback raised an exception: {str(e)}"
                logger.exception(msg)
        return result
    except Exception as e:
        msg = f"{future_name}| raised an exception: {str(e)}"
        logger.exception(msg)

        if score_or_measure is not None:
            score_or_measure.state = TaskState.FAILURE
            score_or_measure.finished = True
            logger.info(f"{future_name}| successfully updated state to failure")
        return None


def safe_name(value: str) -> str:
    return pkg_resources.to_filename(pkg_resources.safe_name(value))


def gudexc(
    msg: str,
    score: Union["Score", str, None] = None,
    player: Union["Player", str, None] = None,
    measure: Union["Measure", str, None] = None,
    host: Union["Host", str, None] = None,
) -> str:
    exc_type, exc_value = sys.exc_info()[:2]
    if exc_type or exc_value:
        msg = f"{msg} ({exc_type.__name__}: {exc_value})"
    return gudlog(msg, score, player, measure, host)


def gudlog(
    msg: str,
    score: Union["Score", str, None] = None,
    player: Union["Player", str, None] = None,
    measure: Union["Measure", str, None] = None,
    host: Union["Host", str, None] = None,
) -> str:
    prefix = gudprefix(score, player, measure, host)
    return f"{prefix}| {msg}"


def gudprefix(
    score: Union["Score", str, None] = None,
    player: Union["Player", str, None] = None,
    measure: Union["Measure", str, None] = None,
    host: Union["Host", str, None] = None,
) -> str:
    prefix = ""
    if score:
        if isinstance(score, str):
            prefix += f".{score}"
        else:
            prefix += f".{score.name}"
    if measure:
        if isinstance(measure, str):
            prefix += f".{measure}"
        else:
            prefix += f".{measure.name}"
    if player:
        if isinstance(player, str):
            prefix += f".{player}"
        else:
            prefix += f".{player.name}"
    if host:
        if isinstance(host, str):
            prefix += f".{host}"
        else:
            prefix += f".{host.name}"

    if not prefix:
        prefix = "?"
    elif prefix and prefix[0] == ".":
        prefix = prefix[1:]

    return prefix


def celery_group_status(group_result: "GroupResult", short: bool = False) -> Dict:
    ret = {}

    if not group_result:
        return ret

    ret["id"] = group_result.id
    ret["state"] = TaskState.PENDING  # will change below
    ret["finished"] = group_result.ready()
    ret["completed_count"] = group_result.completed_count()
    ret["failed_count"] = 0
    ret["status"] = {}

    progress_current = 0
    tasks = {}
    for task in group_result.children:
        task_status = celery_task_status(task, short=short)

        # update overall state
        if task_state_priority(task_status["state"]) > task_state_priority(
            ret["state"]
        ):
            ret["state"] = task_status["state"]

        if task_status["state"] == TaskState.FAILURE:
            ret["failed_count"] += 1
            if "status" in task_status:
                ret["status"][task.task_id] = task_status["status"]
        elif task_status["state"] == TaskState.SUCCESS:
            progress_current += 1
        elif task_status["state"] == TaskState.PROGRESS:
            if (
                "meta" in task_status
                and task_status["meta"]
                and "current" in task_status["meta"]
                and "total" in task_status["meta"]
            ):
                c = task_status["meta"]["current"]
                t = task_status["meta"]["total"]
                cur_fraction = c / t
                if cur_fraction > 1:
                    logger.warning(
                        f"{task_status['name']}: got task status current > total;"
                        " something is weird"
                    )
                else:
                    progress_current += cur_fraction
            else:
                logger.warning(
                    f"{task_status['name']}: task has status of 'progress' but "
                    "missing or improper 'meta' dict "
                )

        tasks[task.task_id] = task_status

    if len(group_result.children) == 0:
        normalized_progress = 0
    else:
        normalized_progress = progress_current / len(group_result.children)

    if group_result.successful():
        ret["state"] = TaskState.SUCCESS
        normalized_progress = len(group_result.children)

    if not short:
        ret["meta"] = {
            "current": normalized_progress,
        }
        ret["tasks"] = tasks

    return ret


def celery_task_status(task: "AsyncResult", short: bool = False) -> Dict:
    task_status = {"name": task.name, "state": task.status, "meta": {}}

    if task.status == TaskState.FAILURE:
        # task.result is probably an exception
        task_status["traceback"] = task.traceback
        if isinstance(task.result, BaseException):
            exc_type = type(task.result)
            exc_type_str = f"{exc_type.__module__}.{exc_type.__name__}"
            task_status["status"] = f"Exception ({exc_type_str}): {str(task.result)}"
        else:
            task_status["status"] = str(task.result)
    elif task.status == TaskState.SUCCESS:
        task_status["result"] = task.result
    elif task.status == TaskState.PROGRESS:
        task_status["meta"] = task.result
    else:
        pass

    if not short:
        task_status["retries"] = task.retries
        task_status["id"] = task.task_id
        if task.date_done is not None:
            task_status["finished_at"] = task.date_done.isoformat()

    return task_status


def create_johann_tarball() -> bool:
    codehash = get_codehash()
    tarball_name = f"johann.{codehash}.tar.gz"
    tarball_process = subprocess.run(
        [
            "tar",
            "-C",
            str(config.SRC_ROOT),
            "-czf",
            tarball_name,
            "--exclude=__pycache__",
            "--exclude=scores",
            "--exclude=minirepo*",
            "--exclude=*.log",
            "--exclude=.[^/]*",
            "./",
        ],
        cwd=str(config.TARBALL_PATH),
    )

    if tarball_process.returncode != 0:
        logger.warning(
            f"Failed to create tarball for current code: {str(tarball_process.stderr)}"
        )
        return False
    else:
        logger.debug(f"Successfully created Johann tarball {tarball_name}")
        return True


def py_to_clistr(script_str: str) -> str:
    exec_str = "exec(%r)" % re.sub("\r\n|\r", "\n", script_str.rstrip())
    return '"%s"' % exec_str.replace('"', r"\"")


def get_attr(attr_name: str) -> Tuple[Any, Optional[str]]:
    try:
        module_name, attribute = attr_name.rsplit(".", 1)
    except ValueError:
        msg = f"invalid attribute: '{attr_name}'; did you specify a module?"
        return None, msg

    if module_name not in sys.modules:
        msg = f"no such module in current context: '{module_name}'"
        return None, msg

    attr = getattr(sys.modules[module_name], attribute, None)

    if not attr:
        msg = f"no such attribute: '{attr_name}'"
        return None, msg
    else:
        return attr, None


def load_plugins():
    discovered_plugins = []
    exclusions = []
    for _, name, _ in pkgutil.iter_modules():
        if not name.startswith("johann_") or name == "johann_main":
            continue
        for p in config.PLUGINS_EXCLUDE:
            if str.endswith(name, p):
                exclusions.append(name)
        discovered_plugins.append(name)

    if discovered_plugins:
        logger.info(f"Found: {', '.join(discovered_plugins)}")
        if exclusions:
            logger.info(f"Excluding: {', '.join(exclusions)}")
        for plugin in discovered_plugins:
            if plugin not in exclusions:
                logger.debug(f"Importing {plugin}")
                importlib.import_module(plugin)
                active_plugins.append(plugin)
    else:
        logger.info("No plugins found")


def celery_registered_tasks() -> List[str]:
    return list(celery_app.tasks.keys())


# Note: descends recursively into lists and dicts
def transform_args(
    score: "Score", measure: "Measure", player: "Player", args: List
) -> List:
    new_args = []
    for a in args:
        if isinstance(a, str):
            new_arg = transform_arg(score, measure, player, a)
            new_args.append(new_arg)
        elif isinstance(a, list):
            new_arg = transform_args(score, measure, player, a)
            new_args.append(new_arg)
        elif isinstance(a, dict):
            new_arg = {}
            for k, v in a:
                new_arg[k] = transform_args(score, measure, player, v)
            new_args.append(new_arg)
        else:
            new_args.append(a)

    return new_args


def parse_special_arg(a: str) -> Tuple[str, List[str]]:
    if not isinstance(a, str):
        raise ValueError("not a string")

    arg_type = re.match(r"johann\.(\w+)", a)
    stored = re.fullmatch(r"johann\.stored((\.\w+(-\w+)*)+)", a)
    rand = re.fullmatch(r"johann\.random\.(\d+)-(\d+)", a)

    if not arg_type:
        raise ValueError("not a special argument")
    elif arg_type.group(1) not in config.SPECIAL_ARG_TYPES:
        raise ValueError(f"unrecognized special argument type '{arg_type.group(1)}'")
    elif stored:
        keyparts = stored.group(1)[1:].split(".")
        if len(keyparts) == 1:
            key = keyparts[0]
            subkey = None
            subsubkey = None
            subsubsubkey = None
        elif len(keyparts) == 2:
            key = keyparts[0]
            subkey = keyparts[1]
            subsubkey = None
            subsubsubkey = None
        elif len(keyparts) == 3:
            key = keyparts[0]
            subkey = keyparts[1]
            subsubkey = keyparts[2]
            subsubsubkey = None
        elif len(keyparts) == 4:
            key = keyparts[0]
            subkey = keyparts[1]
            subsubkey = keyparts[2]
            subsubsubkey = keyparts[3]
        else:
            raise KeyError("too many subkeys")
        return arg_type.group(1), [key, subkey, subsubkey, subsubsubkey]
    elif rand:
        if rand.group(1) > rand.group(2):
            raise ValueError("range's lower bound must be less than its upper bound")
        else:
            return arg_type.group(1), [rand.group(1), rand.group(2)]
    else:
        raise ValueError("invalid special argument")


def transform_arg(score: "Score", measure: "Measure", player: "Player", a: str) -> Any:
    if type(a) == str and a.startswith("johann."):
        try:
            arg_type, arg_split = parse_special_arg(a)
            if arg_type == "stored":
                if measure.lazy_fetch_stored:
                    # add the score_name so the player knows the right URL to use
                    new_a = f"johann.stored.{score.name}.{'.'.join(arg_split)}"
                    return (
                        new_a  # the player will fetch from conductor at measure runtime
                    )
                else:
                    success, msg, code, data = score.fetch_stored_data(
                        *tuple(arg_split)
                    )
                    if not success:
                        raise KeyError(msg)
                    logger.debug(gudlog(msg, score, player, measure))
                    return data
            elif arg_type == "random":
                new_arg = random.randint(int(arg_split[0]), int(arg_split[1]))
                msg = f"'{a}' randomized to {new_arg}"
                logger.debug(gudlog(msg, score, player, measure))
                return new_arg
        except ValueError:
            msg = gudexc(
                f"Failed to transform special argument '{a}'", score, player, measure
            )
            logger.debug(msg)
            raise
    else:
        return a


def resource_listdir_noext(package_name: str, directory: str) -> List[str]:
    """Same as pkg_resources.resource_listdir but removes file extensions.

    Args:
        package_name: The name of the package.
        directory: The name of the directory within the package.

    Returns:
        The resource filenames with their extensions removed.
    """
    resources = pkg_resources.resource_listdir(package_name, directory)
    return [os.path.splitext(x)[0] for x in resources]


def get_score_resources(score_dir: str = "scores") -> Dict[str, List[str]]:
    """Gets the names of scores included in the packages of johann and any active plugins.

    Args:
        score_dir:
            Optional; The name of the package subdirectory in which scores reside.
            Defaults to "scores".

    Returns:
        A dict mapping package names to score resource names.
    """
    score_resources = {"johann": resource_listdir_noext("johann", score_dir)}
    for plugin_name in active_plugins:
        if not pkg_resources.resource_isdir(plugin_name, score_dir):
            logger.debug(f"No '{score_dir}' directory in package {plugin_name}")
            continue
        plugin_scores = resource_listdir_noext(plugin_name, score_dir)
        logger.debug(f"{plugin_name}: found these scores: {plugin_scores}")
        score_resources[plugin_name] = plugin_scores
    return score_resources


def johann_response(
    success: bool,
    msgs: Union[str, List, None] = None,
    status_code: int = 200,
    data: Any = None,
    sort_keys: bool = False,
) -> aiohttp.web.Response:
    if msgs is None:
        msgs = []
    if isinstance(msgs, str):
        msgs = [msgs]

    ret = {"success": success, "messages": msgs, "data": data}

    return aiohttp.web.Response(
        body=json.dumps(ret, indent=4, sort_keys=sort_keys).encode("utf-8"),
        content_type="application/json",
        status=status_code,
    )



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
from enum import Enum


class StrEnum(str, Enum):
    pass


class TaskState(StrEnum):
    # shared with State
    FAILURE = "FAILURE"
    SUCCESS = "SUCCESS"
    PENDING = "PENDING"
    STARTED = "STARTED"
    RETRY = "RETRY"
    PROGRESS = "PROGRESS"
    QUEUED = "QUEUED"
    DEFERRED = "DEFERRED"  # waiting on a dependency before being queued


class PmtrVariant(StrEnum):
    NONE = "NONE"
    DEVUDP = "DEVUDP"  # NOTE: requires bash (e.g., won't work on vanilla Alpine)


class HostOS(StrEnum):
    LINUX = "LINUX"
    WINDOWS = "WINDOWS"
    # DARWIN = 'DARWIN'



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.

import pprint
import subprocess
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from celery.signals import celeryd_init, task_failure, task_retry

import johann.util
from johann.shared.config import JohannConfig, celery_app
from johann.shared.logger import JohannLogger
from johann.tasks_util import RetryableException

try:
    import johann.conductor_tasks  # noqa: F401
except ImportError:
    pass

if TYPE_CHECKING:
    from logging import Logger
    from traceback import StackSummary

    from billiard.einfo import ExceptionInfo
    from celery import Task
    from celery.worker.request import Request


config: Optional[JohannConfig] = None
logger: Optional["Logger"] = None


@celeryd_init.connect
def configure_workers(sender=None, conf=None, **kwar):
    global config
    global logger

    config = JohannConfig.get_config()
    logger = JohannLogger(__name__).logger

    johann.util.load_plugins()
    logger.debug(
        "Celery registered"
        f" tasks:\n{pprint.pformat(johann.util.celery_registered_tasks())}"
    )

    logger.debug(config.json(indent=2, sort_keys=True))


@task_failure.connect
def log_failure(
    sender: Optional["Task"] = None,
    task_id: str = None,
    exception: Exception = None,
    args: Optional[List] = None,
    kwargs: Optional[Dict] = None,
    traceback: Optional["StackSummary"] = None,
    einfo: Optional["ExceptionInfo"] = None,
    **celery_kwargs: Any,
) -> None:
    msg_id = (
        ((sender.request.shadow if sender.request else sender.name) if sender else None)
        or task_id
        or "From unknown task"
    )
    if exception:
        msg = f"{msg_id}: TASK FAILED -- ({type(exception)}): {exception}"
        logger.exception(msg)
    else:
        msg = f"{msg_id}: TASK FAILED -- unknown Exception"
        logger.error(msg)


@task_retry.connect
def log_retry(
    sender: Optional["Task"] = None,
    request: Optional["Request"] = None,
    reason: Optional[Union[Exception, str]] = None,
    einfo: Optional["ExceptionInfo"] = None,
    **celery_kwargs: Any,
) -> None:
    msg_id = (
        (sender.request.shadow if sender.request else sender.name) if sender else None
    ) or "From unknown task"
    if reason:
        if isinstance(reason, Exception):
            logger.exception(f"{msg_id}: TASK RETRY -- {reason}")
        else:
            logger.warning(f"{msg_id}: TASK_RETRY -- {reason}")
    else:
        logger.warning(f"{msg_id}: TASK RETRY -- unknown reason")


@celery_app.task
def run_shell_command(
    command: str, shell: str = "sh", allow_errors: bool = True
) -> str:
    logger.debug(f"Running command '{command}'")
    proc = subprocess.run(
        [shell, "-c", command],
        encoding="utf-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    exit_code = proc.returncode
    output = proc.stdout

    logger.debug(f"Command output:\n{output}")
    if exit_code:
        msg = "Non-zero exit code"
        if not allow_errors:
            logger.error(msg)
            raise Exception(msg)
        else:
            logger.warning(msg)
    return output


@celery_app.task
def pmtr_enable(pmtr_job_name: str) -> str:
    cmd = f"echo -n 'enable {pmtr_job_name}' > /dev/udp/127.0.0.1/31337"
    return run_shell_command(cmd, "bash")


@celery_app.task
def pmtr_disable(pmtr_job_name: str) -> str:
    cmd = f"echo -n 'disable {pmtr_job_name}' > /dev/udp/127.0.0.1/31337"
    return run_shell_command(cmd, "bash")


@celery_app.task(
    bind=True, autoretry_for=[RetryableException], retry_backoff=3, max_retries=3
)
def test_retry(self: "Task") -> None:
    time.sleep(5)
    if self.request.retries > 0:
        logger.info("This is a retry")
        if self.request.retries % 2 == 1:
            logger.info("This is an odd numbered retry -- fail")
            raise RetryableException("purposefully failed")
        else:
            logger.info("This is an even numbered retry -- pass")
            return
    else:
        logger.info("This is not a retry -- fail")
        raise RetryableException("purposefully failed")



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import asyncio
import json
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import marshmallow
import ruamel.yaml
from ruamel.yaml.error import YAMLError, YAMLFutureWarning, YAMLWarning

from johann import util
from johann.host import HostSchema
from johann.player import PlayerSchema
from johann.score import ScoreSchema
from johann.shared.config import JohannConfig, hosts, scores
from johann.shared.logger import JohannLogger

if TYPE_CHECKING:
    from aiohttp.web import Request, Response

    from johann.host import Host
    from johann.measure import Measure
    from johann.player import Player
    from johann.score import Score


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger

yaml = ruamel.yaml.YAML(typ="safe")
yaml.default_flow_style = False


# roll_call and cue_the_music in one call
async def affrettando(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    r = await roll_call(request)
    if r.status >= 300:
        return r

    r = await cue_the_music(request)
    return r


def read_scores() -> None:
    for package_name, score_names in util.get_score_resources("scores").items():
        for s in score_names:
            if not config.TESTING and s.startswith("test"):
                msg = f"Skipping '{s}' from package {package_name}"
                logger.info(msg)
                continue

            read_score(s, package_name)


# read a 'score' (yaml file describing experiment/scenario) from python package resources
def read_score(
    score_name: str,
    package_name: str = None,
    score_dir: str = "scores",
) -> Tuple[Optional[Dict[str, Any]], Optional[str], int]:
    if package_name in config.PLUGINS_EXCLUDE:
        msg = f"Skipping {score_name} due to package '{package_name}' exclusion"
        logger.info(msg)
        return None, msg, 400

    try:
        # util._validate_score_name_dir
        score_str = util.get_score_str(score_name, package_name, score_dir)
        if not score_str:
            return None, "score not found", 404

        score_dict = yaml.load(score_str)
        score: "Score" = ScoreSchema().load(score_dict)
        score.package = package_name
    except (YAMLError, YAMLWarning, YAMLFutureWarning) as e:
        msg = f"score '{score_name}' YAML failed to parse:\n{str(e)}"
        logger.warning(msg)
        return None, msg, 400
    except marshmallow.ValidationError as e:
        msg = f"score '{score_name}' failed validation:\n{str(e)}"
        logger.warning(msg)
        return None, msg, 400
    except OSError as e:
        msg = f"failed to read score '{score_name}': {e.strerror}"
        logger.warning(msg)
        return None, msg, 500
    except Exception as e:
        logger.exception(e)
        msg = f"unexpected error reading score '{score_name}'; see logs for details"
        logger.warning(msg)
        return None, msg, 500

    scores[score.name] = score
    logger.debug(
        f"read score '{score_name}'{' from package {package_name}' if package_name else ''}"
    )

    return score_dict, None, 200


async def api_read_score(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]

    # check for force
    params = request.rel_url.query
    if "force" in params:
        if score_name not in scores:
            logger.warning("Force resetting a score that isn't loaded")
        else:
            del scores[score_name]

    score_dict, err_msg, status_code = read_score(score_name)

    if score_dict:
        return util.johann_response(True, [], data=score_dict, status_code=status_code)
    else:
        return util.johann_response(False, err_msg, status_code=status_code)


async def get_scores(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")

    ret = {}

    for name, score in scores.items():
        if score.category not in ret:
            ret[score.category] = []
        ret[score.category].append({"name": name, "description": score.description})
        ret[score.category].sort(key=lambda t: t["name"])

    ret = OrderedDict(sorted(ret.items(), key=lambda t: t[0]))  # sort by category
    return util.johann_response(True, [], data=ret)


async def get_hosts(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    ret = {}

    for name, host in hosts.items():
        ret[name] = {"name": host.name, "image": host.get_image()}

    return util.johann_response(True, [], data=ret)


async def get_score(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    return util.johann_response(True, [], data=score.dump())


async def get_host(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    host_name = request.match_info["host_name"]
    if host_name not in hosts:
        return util.johann_response(False, f"unrecognized host '{host_name}'", 404)

    host = hosts[host_name]
    return util.johann_response(True, [], data=host.dump())


# read json file providing hosts to seed Johann with (vice adding via /add_hosts or via the GUI)
def read_hosts_file(file_path: Path) -> Tuple[bool, List[str]]:
    file_path: Path = Path(util.safe_name(str(file_path)))
    if not file_path.is_file():
        msg = f"Provided hosts file does not exist: '{file_path}'"
        logger.warning(msg)
        return False, [msg]

    try:
        with file_path.open() as infile:
            hosts_obj = json.load(infile)
            if type(hosts_obj) is not dict or "hosts" not in hosts_obj:
                msg = f"invalid format for hosts file '{file_path}'"
                logger.warning(msg)
                return False, [msg]

            success, err_msgs, successful_hostnames = _update_hosts(
                hosts_obj["hosts"], allow_invalid=True
            )
            if success:
                logger.debug(
                    f"Hosts successfully added/updated from '{file_path}':"
                    f" {successful_hostnames}"
                )
            return success, err_msgs
    except json.JSONDecodeError as e:
        msg = f"invalid JSON in hosts file '{file_path}'"
        logger.warning(e)
        return False, [msg]
    except AttributeError as e:
        msg = "unable to parse hosts file '{}': {}".format(file_path, str(e))
        logger.warning(msg)
        return False, [msg]
    except marshmallow.ValidationError as e:
        msg = "one or more hosts in '{}' failed validation:\n{}".format(
            file_path, str(e)
        )
        logger.warning(msg)
        return False, [msg]
    except Exception as e:
        logger.exception(e)
        msg = "unexpectedly failed to read hosts file '{}'".format(file_path)
        logger.exception(msg)
        return False, [msg]


# assumes some validation has been done prior
# should be called within try/except
def _update_hosts(
    hosts_dict: Dict[str, Any], allow_invalid: bool = False
) -> Tuple[bool, Optional[List[str]], List[str]]:
    valid_hosts = []
    err_msgs = []
    success = True
    for h_name, h_data in hosts_dict.items():
        if "hostname" not in h_data:
            h_data["hostname"] = h_name
        try:
            h: "Host" = HostSchema().load(h_data)
            valid_hosts.append(h)
        except marshmallow.ValidationError as e:
            msg = f"invalid host data provided ({h_name}): {str(e)}"
            logger.warning(msg)
            err_msgs.append(msg)
            success = False

    # if any failures, bail before modifying Hosts
    if not success and not allow_invalid:
        return False, err_msgs, []

    to_remove = []
    for h in valid_hosts:
        if h.name in hosts:
            extant_host: Host = hosts[h.name]
            success, err_msg = extant_host.copy_from(h)
            if success:
                logger.debug(f"updated host '{h.name}' to:\n{extant_host.dump()}")
            else:
                msg = f"failed to update host '{h.name}': {err_msg}"
                logger.warning(msg)
                err_msgs.append(msg)
                to_remove.append(h)
                success = False
        else:
            hosts[h.name] = h
            logger.debug(f"added host '{h.name}'")

    for h in to_remove:
        valid_hosts.remove(h)

    return success, err_msgs, [h.name for h in valid_hosts]


async def add_hosts(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    try:
        data = await request.json()
    except json.JSONDecodeError as e:
        msg = "add_hosts: invalid json"
        logger.warning(f"{msg}\n{str(e)}")
        return util.johann_response(False, msg, 400)

    logger.debug("add_hosts endpoint called")
    if "hosts" in data and isinstance(data["hosts"], dict):
        success, err_msgs, successful_hostnames = _update_hosts(
            data["hosts"], allow_invalid=False
        )

        if not success:
            return util.johann_response(False, err_msgs, 400)
        else:
            return util.johann_response(True, [], data=successful_hostnames)
    else:
        msg = "invalid format for key 'hosts'"
        logger.warning(msg)
        return util.johann_response(False, msg, 400)


async def get_score_raw(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    return util.johann_response(
        True, [], data=score.dump(exclude_local=False, yaml_fields_only=False)
    )


async def get_score_status(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    return util.johann_response(True, [], data=score.get_status())


async def get_score_status_short(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    return util.johann_response(True, [], data=score.get_status(short=True))


async def get_score_status_alt(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    status = score.get_status(short=False)

    ret = {}
    measures = {}
    current = 0
    totals = score.get_host_totals()
    total = totals["total"]
    fails = 0
    for m_name, m_status in status["measures"].items():
        measure_ret = {}
        m_current = 0
        m_fails = 0
        for p_status in m_status["task_status"].values():
            m_fails += p_status["failed_count"]
            m_current += p_status["meta"]["current"]
        measure_ret["total"] = totals[m_name]["total"]
        measure_ret["current"] = m_current
        measure_ret["failed_count"] = m_fails
        current += m_current
        fails += m_fails
        measure_ret["state"] = m_status["state"]
        measures[m_name] = measure_ret

    ret["current"] = current
    ret["total"] = total
    ret["failed_count"] = fails
    ret["state"] = status["state"]
    ret["status"] = ret["state"]
    ret["measures"] = measures
    ret["raw"] = status
    return util.johann_response(True, [], data=ret)


async def get_score_measures(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    return util.johann_response(True, [], data=[x.name for x in score.measures])


def _get_measure_helper(
    request: "Request",
) -> Tuple[Optional["Measure"], Optional[str]]:
    score_name = request.match_info["score_name"]
    measure_name = request.match_info["measure_name"]
    if score_name not in scores:
        return None, f"unrecognized score '{score_name}'"

    score = scores[score_name]
    measure_names = [x.name for x in score.measures]
    if measure_name not in measure_names:
        return None, f"unrecognized measure '{measure_name}'"

    measure = [x for x in score.measures if x.name == measure_name][0]
    return measure, None


async def get_measure(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    measure, msg = _get_measure_helper(request)
    if not measure:
        return util.johann_response(False, msg, 404)

    return util.johann_response(True, [], data=measure.dump())


async def get_measure_status(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    measure, msg = _get_measure_helper(request)
    if not measure:
        return util.johann_response(False, msg, 404)

    return util.johann_response(True, [], data=measure.get_status())


async def manually_play_measure(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    measure_name = request.match_info["measure_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    score = scores[score_name]
    measure_names = [x.name for x in score.measures]
    if measure_name not in measure_names:
        return util.johann_response(
            False, f"unrecognized measure '{measure_name}'", 404
        )

    measure = [x for x in score.measures if x.name == measure_name][0]
    m = measure

    if m.started():
        if (
            "force" in request.rel_url.query
            and request.rel_url.query["force"].lower() != "false"
        ):
            msg = f"Forcing re-play of measure '{m.name}'"
            logger.warning(msg)
        else:
            return util.johann_response(
                False,
                "measure already played/playing; to run anyway, include query param"
                " 'force=true'",
                400,
            )
    else:
        msg = f"Manually playing measure '{m.name}'"
        logger.info(msg)

    score.queue_measure(m)
    return util.johann_response(True, msg)


async def roll_call(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)

    logger.debug(f"{request.method} roll_call for score '{score_name}'")
    score = scores[score_name]

    # make sure score isn't already running (e.g. by another user)
    if score.started_at and not score.finished:
        return util.johann_response(False, "score already playing", 400)

    # update score/players if changed
    if request.method == "POST":
        data = await request.json()
        score.create_hosts = data["create_hosts"]  # PLANNED: REMOVE
        score.discard_hosts = data["discard_hosts"]  # PLANNED: REMOVE

        if "players" in data and isinstance(data["players"], dict):
            posted_players = {}
            for p_name, p_data in data["players"].items():
                try:
                    p: "Player" = PlayerSchema().load(p_data)
                except marshmallow.ValidationError as e:
                    msg = f"invalid player data provided ({p_name}): {str(e)}"
                    logger.warning(f"roll_call: {msg}")
                    return util.johann_response(False, msg, 400)

                posted_players[p.name] = p
                if p.name in score.players:
                    score_player = score.players[p.name]
                    success, err_msg = score_player.copy_from(p, score)
                    if success is None:
                        pass  # no changes required
                    elif success:
                        logger.debug(
                            f"roll_call: updated player '{p.name}'"
                            f" to:\n{score_player.dump()}"
                        )
                    else:
                        msg = f"failed to update player '{p.name}': {err_msg}"
                        logger.warning(f"roll_call: {msg}")
                        return util.johann_response(False, msg, 400)
                else:
                    msg = f"unrecognized player: '{p.name}'"
                    logger.warning(f"roll_call: {msg}")
                    return util.johann_response(False, msg, 400)
        else:
            logger.warning(
                "roll_call: POST missing or invalid format for key 'players'"
            )

    # actual roll call
    success, err_msgs = score.validate_create_host_mappings()
    if not success:
        logger.warning(f"{score_name}: errors validating host mappings:\n{err_msgs}")
        return util.johann_response(False, err_msgs, 400)
    else:
        score.last_successful_roll_call = datetime.utcnow()
        return util.johann_response(
            True, "roll_call successful; you are now free to cue the music"
        )


async def cue_the_music(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)
    score = scores[score_name]

    # make sure score isn't already running (e.g. by another user)
    if score.started_at and not score.finished:
        return util.johann_response(False, "score already playing", 400)

    # make sure each Player in the Score is matched to actual entities/containers
    success, err_msgs = score.validate_create_host_mappings()
    if not success:
        logger.warning(
            f"{score.name}: tuning failed; errors validating host mappings:\n{err_msgs}"
        )
        return util.johann_response(False, err_msgs, 400)

    task = asyncio.ensure_future(score.play())
    task = util.wrap_future(task, score.name, None, score_or_measure=score)
    asyncio.ensure_future(task)

    return util.johann_response(True, "score is playing")


async def retrieve_stored_data_all(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    return await _retrieve_stored_data(score_name)


async def retrieve_stored_data_1(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    key = request.match_info["key"]
    return await _retrieve_stored_data(score_name, key)


async def retrieve_stored_data_2(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    key = request.match_info["key"]
    subkey = request.match_info["subkey"]
    return await _retrieve_stored_data(score_name, key, subkey)


async def retrieve_stored_data_3(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    key = request.match_info["key"]
    subkey = request.match_info["subkey"]
    subsubkey = request.match_info["subsubkey"]
    return await _retrieve_stored_data(score_name, key, subkey, subsubkey)


async def retrieve_stored_data_4(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    score_name = request.match_info["score_name"]
    key = request.match_info["key"]
    subkey = request.match_info["subkey"]
    subsubkey = request.match_info["subsubkey"]
    subsubsubkey = request.match_info["subsubsubkey"]

    # convenience conversion of 'None' or 'none' keys to actual None
    # this is used in tasks_main.py for lazy/J.I.T. fetching
    if key.lower() == "none":
        key = None
    if subkey.lower() == "none":
        subkey = None
    if subsubkey.lower() == "none":
        subsubkey = None
    if subsubsubkey.lower() == "none":
        subsubsubkey = None

    return await _retrieve_stored_data(score_name, key, subkey, subsubkey, subsubsubkey)


async def _retrieve_stored_data(
    score_name: str,
    key: Optional[str] = None,
    subkey: Optional[str] = None,
    subsubkey: Optional[str] = None,
    subsubsubkey: Optional[str] = None,
) -> "Response":
    if score_name not in scores:
        return util.johann_response(False, f"unrecognized score '{score_name}'", 404)
    score = scores[score_name]

    success, msg, code, data = score.fetch_stored_data(
        key, subkey, subsubkey, subsubsubkey
    )
    logger.debug(f"retrieve_stored_data API call: {msg}")
    return util.johann_response(success, msg, code, data=data)


async def api_get_codehash(request: "Request") -> "Response":
    if config.TRACE:
        logger.debug(f"{request.url}")
    return util.johann_response(True, [], data=util.get_codehash())



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import hashlib
import os
import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from johann.shared.config import JohannConfig
from johann.shared.enums import HostOS, PmtrVariant
from johann.shared.logger import JohannLogger
from johann.util import create_johann_tarball, get_codehash, py_to_clistr

if TYPE_CHECKING:
    from johann.host import Host
    from johann.util import PathLikeObj


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


class HostControl(ABC):
    @abstractmethod
    def __init__(self, host_copy: "Host"):
        self.control_method = host_copy.control_method
        self.control_name = host_copy.control_name
        self.name = host_copy.name
        self.os = host_copy.os
        self.pip_offline_install = host_copy.pip_offline_install
        self.pmtr_variant = host_copy.pmtr_variant
        self.pwd_env = host_copy.pwd_env
        self.python_path = host_copy.python_path
        self.python_ver = host_copy.python_ver
        self.user = host_copy.user

    @staticmethod
    @abstractmethod
    def get_host_names() -> List[str]:
        pass

    @staticmethod
    @abstractmethod
    def host_exists(host_name: str) -> bool:
        pass

    @abstractmethod
    def put_archive(
        self,
        archive_path: "PathLikeObj",
        dest_path_inc_filename: "PathLikeObj",
        remove_archive_file: bool = True,
    ) -> bool:
        pass

    @abstractmethod
    def run_cmd(
        self,
        cmd: str,
        environment: Optional[Dict[str, str]] = None,
        detach: bool = False,
        privileged: bool = False,
        workpath: Optional["PathLikeObj"] = None,
        finish_timeout: int = -1,
        strip_output: bool = True,
    ) -> Tuple[Optional[int], Optional[str]]:
        """

        Args:
            cmd:
            environment:
            detach:
            privileged:
            workpath:
            finish_timeout:
            strip_output:

        Returns:
            Tuple of:
                exit code or None
                command output or None

        """
        pass

    def get_python_version(self) -> Tuple[bool, Optional[str]]:
        cmd = (
            f"/bin/sh -ec \"{self.python_path} --version | cut -d ' ' -f2 | cut -d."
            ' -f1-2"'
        )
        exit_code, output = self.run_cmd(cmd)
        if not output:
            return False, None
        else:
            return exit_code == 0, output

    def get_os(self) -> Tuple[bool, Optional[str]]:
        cmd = (
            f'{self.python_path} -c "'
            f"""exec(\\"import platform\\nprint(platform.system())\\")\""""
        )
        exit_code, output = self.run_cmd(cmd)
        if output:
            output = output.upper()

        if not exit_code:
            return True, output
        else:
            return False, output

    # returns None on success, or error str on failure
    def install_pip_dependencies(
        self, workpath: "PathLikeObj", python_verstr: str
    ) -> Tuple[bool, Optional[str]]:
        pip_cmd = f"{self.python_path} -m pip install"
        if self.user:
            pip_cmd += " --user"
        if self.pip_offline_install:
            # offline install
            logger.debug(f"{self.name}: Using offline pip install")
            pip_cmd += f" --no-index -f minirepo -r {str(workpath)}/requirements.txt"
        else:
            # online install:
            logger.debug(f"{self.name}: Using online pip install")
            pip_cmd += (
                f" --index={config.PYPI_INDEX_URL} --retries=2 --timeout=3"
                f" --trusted-host {config.PYPI_INDEX_HOST}"
            )
        pip_cmd += " --disable-pip-version-check"
        # install plugins.txt separately so that it doesn't have to have hashes
        cmd = (
            f'/bin/sh -ec "{pip_cmd} -r {str(workpath)}/requirements.txt && {pip_cmd}'
            f' -r {str(workpath)}/plugins.txt"'
        )

        exit_code, output = self.run_cmd(cmd, workpath=workpath)
        if exit_code:
            errmsg = f"possible issues installing pip requirements: {output}"
            return False, errmsg

        return True, None

    def johann_control(
        self, disable: bool, workpath: "PathLikeObj"
    ) -> Tuple[bool, Optional[str]]:
        environment = None
        if self.pmtr_variant == PmtrVariant.DEVUDP:
            cmd = (
                f"/bin/bash -c \"echo -n '{'disable' if disable else 'enable'}"
                " johann_player' > /dev/udp/127.0.0.1/31337\""
            )
            detach = False
        elif self.pmtr_variant == PmtrVariant.NONE:
            if disable:
                return self.python_pkill("johann_main.py")
            else:
                # cmd = f"CELERY_QUEUE_ID={self.name} {self.python_path} johann_main.py"
                cmd = f"{self.python_path} johann_main.py"
                environment = {
                    "CELERY_DETACH": "True",
                    "PYTHONPATH": config.DEPLOY_PYTHONPATH_POSIX,
                }
                detach = True
        else:
            print(self.pmtr_variant == PmtrVariant.NONE)
            raise NotImplementedError
        exit_code, output = self.run_cmd(
            cmd, environment=environment, detach=detach, workpath=workpath
        )
        if not detach and exit_code:
            errmsg = (
                f"possible issues {'disabling' if disable else 'enabling'} Johann:"
                f" {output}"
            )
            logger.debug(f"{self.name}: {errmsg}")
            return False, errmsg
        return True, ""

    def python_pgrep(
        self, program_str: str, require_output=False
    ) -> Tuple[Optional[bool], Optional[str]]:
        scriptstr = f"""import os
import sys
try:
    import psutil
except ImportError:
    print('psutil not installed yet')
    sys.exit(2)

procs = []
for proc in psutil.process_iter():
    cmd = ' '.join(proc.cmdline())
    if '{program_str}' in cmd and not 'IGNORE_THIS_PROGRAM' in cmd:
        procs.append(proc)

this_proc = [proc for proc in procs if proc.pid == os.getpid()]
for proc in this_proc:
    procs.remove(proc)

if len(procs) > 0:
    for proc in procs:
        print(str(proc.pid) + ': ' + ' '.join(proc.cmdline()))
    sys.exit(0)
else:
    print("no matching processes found")
    sys.exit(1)"""
        cmd = f"{self.python_path} -c {py_to_clistr(scriptstr)}"
        exit_code, output = self.run_cmd(cmd)
        logger.debug(f"{self.name}: pgrep for {program_str}: {output}")

        if exit_code == 2:
            return None, "psutil not installed yet"
        else:
            return exit_code == 0, output

    def python_pkill(
        self, program_str: str, require_output=False
    ) -> Tuple[Optional[bool], Optional[str]]:
        scriptstr = f"""import os
import sys
try:
    import psutil
except ImportError:
    print('psutil not installed yet')
    sys.exit(2)

procs = []
for proc in psutil.process_iter():
    cmd = ' '.join(proc.cmdline())
    if '{program_str}' in cmd and not 'IGNORE_THIS_PROGRAM' in cmd:
        procs.append(proc)

this_proc = [proc for proc in procs if proc.pid == os.getpid()]
for proc in this_proc:
    procs.remove(proc)

if len(procs) > 0:
    print('Matching procs: ' + ', '.join([str(proc.pid) for proc in procs]))
    for proc in procs:
        print('Terminating ' + str(proc.pid) + ': ' + ' '.join(proc.cmdline()))
        try:
            proc.terminate()
        except psutil.NoSuchProcess:
            pass
    gone, alive = psutil.wait_procs(procs, timeout=5)
    if alive:
        print('One or more processes alive after terminate')
        for proc in alive:
            try:
                proc.kill()
            except psutil.NoSuchProcess:
                pass
        gone, alive = psutil.wait_procs(alive, timeout=5)
        if alive:
            print("One or more processes alive after kill; giving up")
            sys.exit(1)
        else:
            print("All matching processes ended after kill")
            sys.exit(0)
    else:
        print("Matching processes successfully terminated")
else:
    print("no matching processes found")
    sys.exit(0)"""
        cmd = f"{self.python_path} -c {py_to_clistr(scriptstr)}"
        exit_code, output = self.run_cmd(cmd, privileged=True)
        logger.debug(f"{self.name}: pkill for {program_str}: {output}")

        if exit_code == 2:
            return None, "psutil not installed yet"
        else:
            return exit_code == 0, output

    def push_johann(self, update_only: bool = False) -> Tuple[bool, Optional[str]]:
        name_is_control = self.control_name == self.name
        control_name = self.control_name or self.name

        if update_only:
            logger.info(
                f"{self.name}: updating Johann"
                f"""{f" using control_name '{control_name}'" if name_is_control else ''}"""
            )
        else:
            logger.info(
                f"{self.name}: installing Johann"
                f"""{f" using control_name '{control_name}'" if name_is_control else ''}"""
            )

        try:
            python_verstr = self.python_ver
            if not python_verstr:
                # make sure we have a supported version of python installed
                success, python_verstr = self.get_python_version()
                if not success:
                    msg = (
                        f"{self.name}: failed to get valid python version; output:"
                        f" {python_verstr}"
                    )
                    logger.warning(msg)
                    return False, msg

            if python_verstr not in config.SUPPORTED_PYTHON_VERSIONS:
                msg = f"{self.name}: invalid python version: {python_verstr}"
                logger.warning(msg)
                return False, msg

            host_os = self.os
            if not self.os:
                # get host os (i.e. 'Linux' or 'Windows')
                success, host_os = self.get_os()
                if not success:
                    msg = f"{self.name}: failed to determine OS and none provided"
                    logger.warning(msg)
                    return False, msg

            if host_os == HostOS.LINUX:
                workpath = config.DEPLOY_PATH_POSIX
                temppath = config.TEMP_PATH_POSIX
            else:
                msg = f"{self.name}: Unsupported OS '{host_os}'"
                logger.warning(msg)
                return False, msg

            logger.debug(
                f"{self.name}: workdir {str(workpath)}; tempdir {str(temppath)}"
            )

            codehash = get_codehash()
            johann_tarball_path = f"{config.TARBALL_PATH}/johann.{codehash}.tar.gz"
            pip_tarball_path = f"{config.TARBALL_PATH}/minirepo.tar.gz"

            # check for tarball for current code
            if not os.path.isfile(johann_tarball_path):
                # create tarball for current code
                if not create_johann_tarball():
                    return False, "Failed to create tarball for current code"
            else:
                logger.debug(
                    f"Conductor (pushing to {self.name}): Johann tarball already"
                    " present locally"
                )

            if not update_only:
                logger.debug(f"{self.name}: Ensuring target directories exist")
                # make sure target workdir exists
                target_dir_1 = str(workpath.joinpath("minirepo"))
                target_dir_2 = str(temppath)
                script_template_str = """import os
os.makedirs('TARGET_DIR_HERE', exist_ok=True)"""
                for target_dir in [target_dir_1, target_dir_2]:
                    script_str = script_template_str.replace(
                        "TARGET_DIR_HERE", target_dir
                    )
                    cmd = f"{self.python_path} -c {py_to_clistr(script_str)}"
                    exit_code, output = self.run_cmd(cmd, detach=False, privileged=True)
                    if exit_code is None or exit_code != 0:
                        msg = f"{self.name}: failed to make dir for Johann: {output}"
                        logger.warning(msg)
                        return False, msg

                # make sure we have permissions on the workpath and minirepo subdir
                if self.user is not None:
                    logger.debug(
                        f"{self.name}: Ensuring user '{self.user}' has ownership of"
                        f" {workpath}"
                    )
                    script_str = f"""import os
import pwd
os.chown('{workpath}', pwd.getpwnam('{self.user}').pw_uid, -1)
os.chown('{workpath.joinpath('minirepo')}', pwd.getpwnam('{self.user}').pw_uid, -1)
                    """
                    cmd = f"{self.python_path} -c {py_to_clistr(script_str)}"
                    exit_code, output = self.run_cmd(cmd, detach=False, privileged=True)
                    if exit_code is None or exit_code != 0:
                        msg = (
                            f"{self.name}: failed to set dir permissions for Johann"
                            f" workpath: {output}"
                        )
                        logger.warning(msg)
                        return False, msg

            johann_was_running, pgrep_output = self.python_pgrep("johann_main.py")
            if johann_was_running:
                logger.debug(f"{self.name}: Johann is running")

            # disable/kill Johann
            if johann_was_running or self.pmtr_variant not in [PmtrVariant.NONE]:
                self.johann_control(True, workpath)
                logger.debug(f"{self.name}: giving johann a couple seconds to close")
                time.sleep(3)

            # kill workers
            self.python_pkill(f"celery -A {config.CELERY_TASKS_MODULE}")

            # push Johann tarball to host
            logger.debug(f"{self.name}: pushing Johann tarball")
            dest_path_inc_filename = workpath.joinpath(
                os.path.basename(johann_tarball_path)
            )
            success = self.put_archive(johann_tarball_path, str(dest_path_inc_filename))
            if not success:
                msg = f"{self.name}: pushing of Johann tarball failed"
                logger.warning(msg)
                return False, msg

            # doing docker commands too close together seems to cause issues
            time.sleep(1)

            # dev versions of pmtr conf's have 'depends' blocks
            # and will restart Johann when the file changes (Johann tarball)
            if config.DEBUG and self.pmtr_variant not in [PmtrVariant.NONE]:
                logger.debug(
                    f"{self.name}: disabling Johann again because of debug mode/pmtr"
                    " depends/johann tarball issue"
                )
                self.johann_control(True, workpath)
                logger.debug(f"{self.name}: giving johann a couple seconds to close")
                time.sleep(3)

            # install Johann requirements if needed
            if not update_only:
                if self.pip_offline_install:
                    dest_path_inc_filename = workpath.joinpath(
                        "minirepo", os.path.basename(pip_tarball_path)
                    )

                    # get hash of pip tarball
                    with open(pip_tarball_path, "rb") as f:
                        md5 = hashlib.md5()
                        for chunk in iter(lambda: f.read(4096), b""):
                            md5.update(chunk)
                        pip_hash = md5.hexdigest()
                        logger.debug(
                            f"{self.name}: Local player pip tarball hash: {pip_hash}"
                        )
                    script_str = f"""import hashlib
with open("{dest_path_inc_filename}", 'rb') as f:
    md5 = hashlib.md5()
    for chunk in iter(lambda:f.read(4096), b""):
        md5.update(chunk)
    print(md5.hexdigest())
                    """
                    cmd = f"{self.python_path} -c {py_to_clistr(script_str)}"
                    exit_code, output = self.run_cmd(cmd)
                    if exit_code == 0 and output == pip_hash:
                        logger.info(
                            f"{self.name}: Player pip tarball already present with"
                            f" matching hash {pip_hash}"
                        )
                    else:
                        if exit_code == 0:
                            logger.debug(
                                f"{self.name}: Player pip tarball hash mismatch:"
                                f" {output} instead of {pip_hash}"
                            )
                        # push pip tarball to host
                        logger.info(
                            f"{self.name}: pushing Player pip tarball; this can take"
                            " some time"
                        )
                        success = self.put_archive(
                            pip_tarball_path,
                            str(dest_path_inc_filename),
                            remove_archive_file=False,
                        )
                        if not success:
                            msg = f"{self.name}: pushing of Player Pip tarball failed"
                            logger.warning(msg)
                            return False, msg

                        # doing commands too close together seems to cause Docker issues
                        time.sleep(1)

                # install dependencies with pip
                success, errmsg = self.install_pip_dependencies(workpath, python_verstr)
                if not success:
                    logger.warning(f"{self.name}: {errmsg}")
                    return False, errmsg

                # doing docker commands too close together seems to cause issues
                time.sleep(1)

            # make sure Johann is dead
            running, pgrep_output = self.python_pgrep("johann_main.py")
            if not update_only and running is None:
                msg = (
                    f"{self.name}: psutil package is still not installed; something is"
                    f" wrong ('{pgrep_output}')"
                )
                logger.warning(msg)
                return False, msg
            elif johann_was_running and running:
                msg = (
                    f"{self.name}: Johann is still running after disable/kill;"
                    " something is wrong"
                )
                logger.warning(msg)
                return False, msg

            # generate and push a .env file with host-specific configuration information
            conf = {
                "CELERY_QUEUE_ID": self.name,
                "DEBUG": 1,
            }
            if self.user:  # don't want to store 'None' string
                conf["CELERY_USER"] = self.user
            # PLANNED:
            # if config.PLUGINS_EXCLUDE:
            #     conf["PLUGINS_EXCLUDE"] = json.dumps(config.PLUGINS_EXCLUDE)
            try:
                if config.HOST_CONTROL_EXTERNAL_REDIS[self.control_method]:
                    conf["REDIS_HOST"] = config.REDIS_HOST_EXTERNAL
            except KeyError:
                pass
            conf_str = "\n".join([f"{k}={v}" for k, v in conf.items()])
            if self.os == HostOS.WINDOWS:
                raise NotImplementedError()
            cmd = (  # sh's echo doesn't need -e
                f"/bin/sh -c \"echo '{conf_str}' > .env\""
            )
            exit_code, output = self.run_cmd(cmd, workpath=workpath)
            if exit_code:
                msg = f"{self.name}: writing .env for johann player failed: {output}"
                logger.warning(msg)
                return False, msg

            # enable/start Johann
            self.johann_control(False, workpath)

            # give Johann some time to start
            count = 15
            interval = 2
            logger.debug(
                f"{self.name}: Waiting up to up to {count} seconds for Johann to start"
            )
            for _ in range(int(count / interval)):
                time.sleep(interval)
                started, pgrep_output = self.python_pgrep("johann_main.py")
                if started:
                    break

            started, pgrep_output = self.python_pgrep("johann_main.py")
            if not started:
                msg = f"{self.name} had issues starting Johann: {pgrep_output}"
                logger.warning(msg)
                return False, msg
            else:
                logger.debug(f"{self.name}: Johann re-enabled")
                return True, None

        except Exception as e:
            msg = f"{self.name}: unexpected exception while pushing Johann ({type(e)})"
            logger.exception(msg)
            return False, msg



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
from pydantic import _PathValueError


class PathNotSafeError(_PathValueError):
    code = "path.not_safe"
    msg_template = 'path "{path}" is not safe'


class JohannError(Exception):
    def __init__(self, msg):
        self.msg = msg



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import random
import time
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import requests

from johann.shared.config import JohannConfig, celery_app
from johann.shared.enums import TaskState
from johann.shared.logger import JohannLogger
from johann.util import get_attr, get_codehash

if TYPE_CHECKING:
    from celery import Task
    from celery.worker.request import Request


config = JohannConfig.get_config()
logger = JohannLogger(__name__).logger


# used when you want to retry a Task
# inherits from BaseException so not caught by 'except Exception'
# but rather is passed to Celery who can then retry
class RetryableException(BaseException):
    pass


@celery_app.task()
def remote_codehash() -> str:
    return get_codehash()


def get_hostname(celery_request: "Request") -> Optional[str]:
    ret = celery_request.hostname
    if "celery@" in ret:
        ret = celery_request.hostname.split("celery@")[1]
        if "_" in ret:
            return ret.split("_")[0]

    return None


@celery_app.task(bind=True)
def select_random(
    self: "Task", select_from: Union[List, Dict], num_select: int, values_only=False
) -> Dict[int, Any]:
    assert type(select_from) in [list, dict]

    ret = []
    indexes_used = []
    rand = random.Random()
    while len(ret) < num_select:
        if len(indexes_used) == len(select_from):
            raise Exception(
                f"{self.request.shadow}: less than requested {num_select} unique items"
                " to choose from"
            )

        index = rand.randint(0, len(select_from) - 1)
        if index in indexes_used:
            continue

        indexes_used.append(index)
        item = (
            list(select_from.values())[index]
            if type(select_from) is dict
            else select_from[index]
        )
        if item not in ret:
            ret.append(item)

    dict_ret = {index: selection for index, selection in enumerate(ret)}
    if num_select == 1:
        ret = list(dict_ret.values())[0]
    elif values_only:
        ret = list(dict_ret.values())
    else:
        ret = dict_ret

    return ret


# assumes arg is str of form 'johann.stored.<SCORE_NAME>[.<SOMETHING>]'
# note that the johann conductor will add the <SCORE_NAME> portion for you
def fetch_stored_data_helper(arg: str) -> Any:
    logger.debug(f"fetching stored_data '{arg}' from conductor")

    keyparts = arg.split("johann.stored.")[1].split(".")
    if len(keyparts) == 1:
        key = "None"
        subkey = "None"
        subsubkey = "None"
        subsubsubkey = "None"
    elif len(keyparts) == 2:
        key = keyparts[1]
        subkey = "None"
        subsubkey = "None"
        subsubsubkey = "None"
    elif len(keyparts) == 3:
        key = keyparts[1]
        subkey = keyparts[2]
        subsubkey = "None"
        subsubsubkey = "None"
    elif len(keyparts) == 4:
        key = keyparts[1]
        subkey = keyparts[2]
        subsubkey = keyparts[3]
        subsubsubkey = "None"
    elif len(keyparts) == 5:
        key = keyparts[1]
        subkey = keyparts[2]
        subsubkey = keyparts[3]
        subsubsubkey = keyparts[4]
    else:
        raise Exception(f"arg {arg} has too many subkeys")
    rjson = requests.get(
        f"http://{config.CONDUCTOR_LOCAL_HOST_NAME}:{config.CONDUCTOR_PORT}/scores/"
        f"{keyparts[0]}/stored_data/{key}/{subkey}/{subsubkey}/{subsubsubkey}"
    ).json()

    if not rjson["success"]:
        raise Exception(f"failed to fetch value for {arg}:\n{rjson['messages']}")
    else:
        data = rjson["data"]
        logger.debug(f"fetched {arg} as:\n{data}")
        return data


# decorator
# fetches any 'johann.stored.<SOMETHING>' args or kwargs to func
def resolve_johann_args(func: Callable) -> Callable:
    @wraps(func)
    def func_wrapper(self, *args: Any, **kwargs: Any) -> Callable:
        new_args = ()
        for arg in args:
            if type(arg) is not str or "johann.stored." not in arg:
                logger.debug(f"ignoring arg {arg}")
                new_args += (arg,)
            elif len(arg.split("johann.stored.")) != 2:
                raise Exception(f"arg {arg} is malformed")
            else:
                logger.debug(f"fetching arg {arg}")
                new_args += (fetch_stored_data_helper(arg),)
        new_kwargs = {}
        for kw, arg in kwargs.items():
            if type(arg) is not str or "johann.stored." not in arg:
                logger.debug(f"ignoring kwarg {arg}")
                new_kwargs[kw] = arg
            elif len(arg.split("johann.stored.")) != 2:
                raise Exception(f"kwarg {kw}={arg} is malformed")
            else:
                logger.debug(f"fetching kwarg {arg}")
                new_kwargs[kw] = fetch_stored_data_helper(arg)

        return func(self, *new_args, **new_kwargs)

    return func_wrapper


@celery_app.task(bind=True)
def do_until(
    self: "Task",
    func_name: str,
    args: Optional[List],
    interval: int,
    timeout: int,
    expected_value: Any = None,
    fetch_inner_func_args: bool = True,
) -> Dict[str, Any]:
    func = get_func_helper(func_name, self)

    start = time.time()
    now = start
    self.update_state(
        state=TaskState.PROGRESS, meta={"current": now - start, "total": timeout}
    )
    result = None
    while (now - start) < timeout:
        time.sleep(interval)
        # if fetch_inner_func_args:
        #    result = resolve_johann_args(func)(self, *args)
        # else:
        if args is None:
            args = ()
        result = func(*args)
        if expected_value is not None and result == expected_value:
            return {
                "current": timeout,
                "total": timeout,
                "status": "Task completed!",
                "result": result,
            }
        else:
            now = time.time()
            self.update_state(
                state=TaskState.PROGRESS,
                meta={
                    "current": now - start,
                    "total": timeout,
                    "interim_result": result,
                },
            )

    if expected_value is not None:
        msg = (
            f"{self.request.shadow}: Timed out waiting for '{expected_value}' from"
            f"{func_name}; last result: {result}"
        )
        logger.warning(msg)
        raise Exception(msg)
    else:
        return {
            "current": timeout,
            "total": timeout,
            "status": "Task completed!",
            "result": result,
        }


def get_func_helper(func_name: str, celery_task: "Task") -> Callable:
    func, msg = get_attr(func_name)

    if not func:
        msg = f"{celery_task.name}: {msg}"
        logger.error(msg)
        celery_task.max_retries = 0
        raise Exception(msg)
    else:
        return func


@celery_app.task(bind=True)
def do_on_random_interval(
    self: "Task",
    func_name: str,
    args: Optional[List],
    interval_min: int,
    interval_max: int,
    duration: int,
) -> Optional[Dict[str, Any]]:
    func = get_func_helper(func_name, self)

    if args is None:
        args = ()

    start = time.time()
    now = start
    self.update_state(
        state=TaskState.PROGRESS, meta={"current": now - start, "total": duration}
    )
    count = 0
    result = None
    while (now - start) < duration:
        count += 1
        sleeptime = random.random() * (interval_max - interval_min) + interval_min
        time.sleep(sleeptime)
        logger.info(
            f"{self.request.shadow}: Calling {func_name} after sleeping {sleeptime}"
        )

        # result = func(*args, **kwargs)
        result = func(*args)
        now = time.time()
        self.update_state(
            state=TaskState.PROGRESS, meta={"current": now - start, "total": duration}
        )
    logger.info(f"{self.request.shadow}: Finished calling {func_name}")

    return {
        "current": duration,
        "total": duration,
        "status": "Task completed!",
        "result": result,
    }



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="johann",
    description="Johann, lightweight and flexible scenario orchestration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jeffrey James",
    author_email="lobotmcj@gmail.com",
    license="BSD-3-Clause",
    url="https://github.com/johannsdg/johann",
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.6, <4",
    install_requires=[
        "aiohttp>=3.6,<4.0",
        "celery==4.4.6",  # 4.4.7 breaks imports somehow
        "dataclasses<0.7",
        "docker>=4.1,<5.0",
        "fastapi>=0.56.0,<1.0",
        "logzero>=1.5,<2.0",
        "marshmallow>=3.2,<4.0",
        "marshmallow_enum>=1.5.0,<2.0",
        "natsort>=6.0.0,<7.0",
        "psutil>=5.6,<6.0",
        "python-dotenv>=0.10.1,<0.11.0",
        "redis>=3.2,<4.0",
        "requests>=2.24.0,<3.0",
        "ruamel.yaml>0.15.0,<0.16",
    ],
)



# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.
from logging import Logger

import logzero
from pydantic import BaseModel, ByteSize

from johann.shared.config import JohannConfig


class LoggerModel(BaseModel):
    config: JohannConfig = JohannConfig.get_config()
    name: str
    file: str = config.LOG_FILE
    level: int = config.LOG_LEVEL
    max_bytes: ByteSize = config.LOG_MAX_BYTES
    backup_count: int = config.LOG_BACKUP_COUNT
    file_level: int = config.LOG_FILE_LEVEL
    format: str = config.LOG_FORMAT


class JohannLogger:
    def __init__(self, name, **kwargs):
        self.model = LoggerModel(name=name, **kwargs)
        self.logger: Logger = logzero.setup_logger(
            name=self.model.name,
            logfile=self.model.file,
            level=self.model.level,
            maxBytes=self.model.max_bytes,
            backupCount=self.model.backup_count,
            fileLoglevel=self.model.file_level,
            formatter=logzero.LogFormatter(fmt=self.model.format),
        )
